from __future__ import annotations

from collections.abc import Iterator, Sequence
from typing import (
    Any,
    Callable,
    cast,
)

import xarray as xr

from muse.agents import AbstractAgent
from muse.production import PRODUCTION_SIGNATURE
from muse.readers.toml import read_technodata
from muse.sectors.abstract import AbstractSector
from muse.sectors.register import register_sector
from muse.sectors.subsector import Subsector
from muse.timeslices import compress_timeslice, expand_timeslice, get_level


@register_sector(name="default")
class Sector(AbstractSector):  # type: ignore
    """Base class for all sectors."""

    @classmethod
    def factory(cls, name: str, settings: Any) -> Sector:
        from muse.interactions import factory as interaction_factory
        from muse.outputs.sector import factory as ofactory
        from muse.production import factory as pfactory

        # Read sector settings
        sector_settings = getattr(settings.sectors, name)._asdict()

        # Extract required settings
        subsectors = sector_settings.get("subsectors")
        if not subsectors:
            raise RuntimeError(f"Missing 'subsectors' section in sector {name}")
        if len(subsectors._asdict()) == 0:
            raise RuntimeError(f"Empty 'subsectors' section in sector {name}")
        interpolation_mode = sector_settings.get("interpolation", "linear")
        timeslice_level = sector_settings.get("timeslice_level", None)
        dispatch_production = sector_settings.get("dispatch_production", "share")
        outputs_config = sector_settings.get("outputs", [])
        interactions_config = sector_settings.get("interactions", None)

        # Read technologies
        technologies = read_technodata(
            settings,
            name,
            interpolation_mode=interpolation_mode,
        )

        # Create subsectors
        subsectors = [
            Subsector.factory(
                subsec_settings,
                technologies,
                regions=settings.regions,
                current_year=int(min(settings.time_framework)),
                name=subsec_name,
                timeslice_level=timeslice_level,
            )
            for subsec_name, subsec_settings in subsectors._asdict().items()
        ]

        # Check that subsector commodities are disjoint
        sector_commodities = [c for s in subsectors for c in s.commodities]
        duplicates = [
            c for c in set(sector_commodities) if sector_commodities.count(c) > 1
        ]
        if duplicates:
            raise RuntimeError(
                f"Commodities {duplicates} are outputted by multiple subsectors."
            )

        # Create outputs
        outputs = ofactory(*outputs_config, sector_name=name)

        # Create production method
        production = pfactory(dispatch_production)

        # Create interactions
        interactions = interaction_factory(interactions_config)

        # Create sector
        return cls(
            name,
            technologies,
            supply_prod=production,
            subsectors=subsectors,
            commodities=sector_commodities,
            outputs=outputs,
            interactions=interactions,
            timeslice_level=timeslice_level,
        )

    def __init__(
        self,
        name: str,
        technologies: xr.Dataset,
        supply_prod: PRODUCTION_SIGNATURE,
        subsectors: Sequence[Subsector] = [],
        commodities: list[str] = [],
        interactions: Callable[[Sequence[AbstractAgent]], None] | None = None,
        outputs: Callable | None = None,
        timeslice_level: str | None = None,
    ):
        from muse.interactions import factory as interaction_factory
        from muse.outputs.sector import factory as ofactory
        from muse.timeslices import TIMESLICE

        """Name of the sector."""
        self.name: str = name

        """Timeslice level for the sector (e.g. "month")."""
        self.timeslice_level = timeslice_level or get_level(TIMESLICE)

        """Subsectors controlled by this object."""
        self.subsectors: Sequence[Subsector] = list(subsectors)

        """Parameters describing the sector's technologies."""
        self.technologies: xr.Dataset = technologies
        if "timeslice" in self.technologies.dims:
            if not get_level(self.technologies) == self.timeslice_level:
                raise ValueError(
                    f"Technodata for {self.name} sector does not match "
                    "the specified timeslice level for that sector "
                    f"({self.timeslice_level})"
                )

        """Interactions between agents.

        Called right before computing new investments, this function should manage any
        interactions between agents, e.g. passing assets from *new* agents  to *retro*
        agents, and maket make-up from *retro* to *new*.

        Defaults to doing nothing.

        The function takes the sequence of agents as input, and returns nothing. It is
        expected to modify the agents in-place.

        See Also
        --------

        :py:mod:`muse.interactions` contains MUSE's base interactions
        """
        self.interactions = interactions or interaction_factory()

        """A function for outputting data for post-mortem analysis."""
        self.outputs: Callable = (
            cast(Callable, ofactory()) if outputs is None else outputs
        )

        """Computes production as used to return the supply to the MCA.

        It can be anything registered with
        :py:func:`@register_production<muse.production.register_production>`.
        """
        self.supply_prod = supply_prod

        """Full supply, consumption and costs data for the most recent year."""
        self.output_data: xr.Dataset

        """Commodities that the sector is in charge of producing."""
        self.commodities: list[str] = commodities

    def next(
        self,
        mca_market: xr.Dataset,
    ) -> xr.Dataset:
        """Advance sector by one time period.

        Args:
            mca_market:
                Market with ``demand``, ``supply``, and ``prices``.

        Returns:
            A market containing the ``supply`` offered by the sector, it's attendant
            ``consumption`` of fuels and materials and the associated ``costs``.
        """
        from logging import getLogger

        def group_assets(x: xr.DataArray) -> xr.DataArray:
            return xr.Dataset(dict(x=x)).groupby("region").sum("asset").x

        # Time period from the market object
        assert len(mca_market.year) == 2
        current_year, investment_year = map(int, mca_market.year.values)
        getLogger(__name__).info(
            f"Running {self.name} for years {current_year} to {investment_year}"
        )

        # Agent interactions
        self.interactions(list(self.agents))

        # Convert market to sector timeslicing
        mca_market = self.convert_to_sector_timeslicing(mca_market)

        # Select appropriate data from the market
        market = mca_market.sel(
            commodity=self.technologies.commodity.values,
            region=self.technologies.region,
        )

        # Select technology data from the investment year
        techs = self.technologies.sel(year=investment_year, drop=True)

        # Perform investments
        for subsector in self.subsectors:
            subsector.invest(technologies=techs, market=market)

        # Full output data
        supply, consume, costs = self.market_variables(market, self.technologies)
        self.output_data = xr.Dataset(
            dict(
                supply=supply,
                consumption=consume,
                costs=costs,
            )
        )

        # Output data for MCA (aggregated over assets)
        if len(supply.region.dims) == 0:
            output_data = self.output_data.sum("asset")
            output_data = output_data.expand_dims(region=[output_data.region.values])
        else:
            output_data = xr.Dataset(
                dict(
                    supply=group_assets(supply),
                    consumption=group_assets(consume),
                    costs=costs,
                )
            )

        result = output_data.copy(deep=True)
        if "dst_region" in result:
            exclude = ["dst_region", "commodity", "year", "timeslice"]
            prices = market.prices.expand_dims(dst_region=market.prices.region.values)
            sup, prices = xr.broadcast(result.supply, prices)
            sup = sup.fillna(0.0)
            con, prices = xr.broadcast(result.consumption, prices)
            con = con.fillna(0.0)
            supply = result.supply.sum("region").rename(dst_region="region")
            consumption = con.sum("dst_region")
            assert len(supply.region) == len(consumption.region)

            # Need to reindex costs to avoid nans for non-producing regions
            costs0, prices = xr.broadcast(result.costs, prices, exclude=exclude)
            # Fulfil nans with price values
            costs0 = costs0.reindex_like(prices).fillna(prices)
            costs0 = costs0.where(costs0 > 0, prices)
            # Find where sup >0 (exporter)
            # Importers have nans and average over exporting price
            costs = ((costs0 * sup) / sup.sum("dst_region")).fillna(
                costs0.mean("region")
            )

            # Take average over dst regions
            costs = costs.where(costs > 0, prices).mean("dst_region")

            result = xr.Dataset(
                dict(supply=supply, consumption=consumption, costs=costs)
            )

        # Convert result to global timeslicing scheme
        return self.convert_to_global_timeslicing(result)

    def save_outputs(self, year: int) -> None:
        """Calls the outputs function with the current output data."""
        self.outputs(self.output_data, self.capacity, year=year)

    def market_variables(self, market: xr.Dataset, technologies: xr.Dataset) -> Any:
        """Computes resulting market: production, consumption, and costs."""
        from muse.costs import levelized_cost_of_energy, supply_cost
        from muse.quantities import capacity_to_service_demand, consumption
        from muse.utilities import broadcast_over_assets, interpolate_capacity

        years = market.year.values
        capacity = interpolate_capacity(self.capacity, year=years)

        # Select technology data for each asset
        # Each asset uses the technology data from the year it was installed
        technodata = broadcast_over_assets(
            technologies, capacity, installed_as_year=True
        )

        # Select relevant investment year prices for each asset
        prices = broadcast_over_assets(market.prices.isel(year=1), capacity)

        # Calculate supply
        supply = self.supply_prod(
            market=market,
            capacity=capacity,
            technologies=technodata,
            timeslice_level=self.timeslice_level,
        )

        # Calculate consumption
        consume = consumption(
            technologies=technodata,
            production=supply,
            prices=prices,
            timeslice_level=self.timeslice_level,
        )

        # Calculate LCOE
        # We select data for the second year, which corresponds to the investment year
        # We base LCOE only on the portion of capacity that is actually used (#728)
        utilized_capacity = capacity_to_service_demand(
            demand=supply.isel(year=1),
            technologies=technodata,
            timeslice_level=self.timeslice_level,
        )
        lcoe = levelized_cost_of_energy(
            prices=prices,
            technologies=technodata,
            capacity=utilized_capacity,
            production=supply.isel(year=1),
            consumption=consume.isel(year=1),
            method="annual",
        )

        # Calculate new commodity prices
        costs = supply_cost(supply, lcoe, asset_dim="asset")

        return supply, consume, costs

    def convert_to_sector_timeslicing(self, market: xr.Dataset) -> xr.Dataset:
        """Converts market data to sector timeslicing."""
        supply = compress_timeslice(
            market["supply"], level=self.timeslice_level, operation="sum"
        )
        consumption = compress_timeslice(
            market["consumption"], level=self.timeslice_level, operation="sum"
        )
        prices = compress_timeslice(
            market["prices"], level=self.timeslice_level, operation="mean"
        )
        return xr.Dataset(dict(supply=supply, consumption=consumption, prices=prices))

    def convert_to_global_timeslicing(self, market: xr.Dataset) -> xr.Dataset:
        """Converts market data to global timeslicing."""
        supply = expand_timeslice(market["supply"], operation="distribute")
        consumption = expand_timeslice(market["consumption"], operation="distribute")
        costs = expand_timeslice(market["costs"], operation="broadcast")
        return xr.Dataset(dict(supply=supply, consumption=consumption, costs=costs))

    @property
    def capacity(self) -> xr.DataArray:
        """Aggregates capacity across agents.

        The capacities are aggregated leaving only two
        dimensions: asset (technology, installation date,
        region), year.
        """
        from muse.utilities import interpolate_capacity, reduce_assets

        traded = [
            u.assets.capacity
            for u in self.agents
            if "dst_region" in u.assets.capacity.dims
        ]
        nontraded = [
            u.assets.capacity
            for u in self.agents
            if "dst_region" not in u.assets.capacity.dims
        ]

        # Only nontraded assets
        if not traded:
            full_list = [
                list(nontraded[i].year.values)
                for i in range(len(nontraded))
                if "year" in nontraded[i].dims
            ]
            flat_list = [item for sublist in full_list for item in sublist]
            years = sorted(list(set(flat_list)))
            capacity = reduce_assets(
                [
                    u.assets.capacity
                    for u in self.agents
                    if "dst_region" not in u.assets.capacity.dims
                ]
            )
            return interpolate_capacity(capacity, year=years)

        # Only traded assets
        elif not nontraded:
            full_list = [
                list(traded[i].year.values)
                for i in range(len(traded))
                if "year" in traded[i].dims
            ]
            flat_list = [item for sublist in full_list for item in sublist]
            years = sorted(list(set(flat_list)))
            capacity = reduce_assets(
                [
                    u.assets.capacity
                    for u in self.agents
                    if "dst_region" in u.assets.capacity.dims
                ]
            )
            return interpolate_capacity(capacity, year=years)

        # Both traded and nontraded assets
        else:
            traded_results = reduce_assets(traded)
            nontraded_results = reduce_assets(nontraded)
            capacity = reduce_assets(
                [
                    traded_results,
                    nontraded_results
                    * (nontraded_results.region == traded_results.dst_region),
                ]
            )
            return interpolate_capacity(capacity, year=years)

    @property
    def agents(self) -> Iterator[AbstractAgent]:
        """Iterator over all agents in the sector."""
        for subsector in self.subsectors:
            yield from subsector.agents



from __future__ import annotations

from collections.abc import MutableMapping, Sequence
from logging import getLogger
from typing import Callable

import numpy as np
import xarray as xr
from scipy.optimize import curve_fit

from muse.mca import FindEquilibriumResults
from muse.registration import registrator

CARBON_BUDGET_METHODS_SIGNATURE = Callable[
    [xr.Dataset, Callable, xr.DataArray, list], float
]
"""carbon budget fitters signature."""

CARBON_BUDGET_FITTERS_SIGNATURE = Callable[[np.ndarray, np.ndarray, int], float]
"""carbon budget fitters signature."""


CARBON_BUDGET_METHODS: MutableMapping[str, CARBON_BUDGET_METHODS_SIGNATURE] = {}
"""Dictionary of carbon budget methods checks."""

CARBON_BUDGET_FITTERS: MutableMapping[str, CARBON_BUDGET_FITTERS_SIGNATURE] = {}
"""Dictionary of carbon budget fitters."""


@registrator(registry=CARBON_BUDGET_FITTERS, loglevel="debug")
def register_carbon_budget_fitter(function: CARBON_BUDGET_FITTERS_SIGNATURE = None):
    """Decorator to register a carbon budget function."""
    return function


@registrator(registry=CARBON_BUDGET_METHODS, loglevel="debug")
def register_carbon_budget_method(function: CARBON_BUDGET_METHODS_SIGNATURE = None):
    """Decorator to register a carbon budget function."""
    return function


def update_carbon_budget(
    carbon_budget: Sequence[float],
    emissions: float,
    year_idx: int,
    over: bool = True,
    under: bool = True,
) -> float:
    """Adjust the carbon budget in the far future if emissions too high or low.

    This feature can allow to simulate overshoot shifting.

    Arguments:
        carbon_budget: budget for future year,
        emissions: emission for future year,
        year_idx: index of year for estimation,
        over: if True, allows overshoot,
        under: if True, allows undershoot.

    Returns:
        An adjusted threshold for the future year
    """
    delta = emissions - carbon_budget[year_idx + 1]
    new_future_budget = carbon_budget[year_idx + 2]
    if over:
        new_future_budget -= max(delta, 0)
    if under:
        new_future_budget -= min(delta, 0)

    return new_future_budget


@register_carbon_budget_method
def fitting(
    market: xr.Dataset,
    equilibrium: Callable[[xr.Dataset], FindEquilibriumResults],
    carbon_budget: xr.DataArray,
    commodities: list,
    refine_price: bool = False,
    price_too_high_threshold: float = 10,
    sample_size: int = 5,
    fitter: str = "linear",
    resolution: int = 2,
) -> float:
    """Used to solve the carbon market.

    Given the emission of a period, adjusts carbon price to meet the budget. A carbon
    market is meant as a pool of emissions for all the modelled regions; therefore, the
    carbon price applies to all modelled regions. The method solves an equation applying
    a fitting of the emission-carbon price relation.

    Arguments:
        market: Market, with the prices, supply, and consumption
        equilibrium: Method for searching market equilibrium
        carbon_budget: limit on emissions
        commodities: list of commodities to limit (ie. emissions)
        refine_price: Boolean to decide on whether carbon price should be capped, with
            the upper bound given by price_too_high_threshold
        price_too_high_threshold: threshold on carbon price
        sample_size: sample size for fitting
        fitter: method to fit emissions with carbon price
        resolution: Number of decimal places to solve the carbon price to

    Returns:
        Adjusted carbon price to meet budget
    """
    # Calculate the carbon price and emissions threshold in the investment year
    future = market.year[-1]
    threshold = carbon_budget.sel(year=future).values.item()
    price = market.prices.sel(year=future, commodity=commodities).mean().values.item()

    # Solve market with current carbon price
    emissions = solve_market(market, equilibrium, commodities, price)

    # Create a sample of prices at which we want to calculate emissions
    sample_prices = create_sample(price, emissions, threshold, sample_size)
    sample_emissions = np.zeros_like(sample_prices)
    sample_emissions[0] = emissions

    # For each sample price, we calculate the new emissions
    for i, new_price in enumerate(sample_prices[1:]):
        sample_emissions[i + 1] = solve_market(
            market, equilibrium, commodities, new_price
        )

    # Based on these results, we finally adjust the carbon price
    new_price = CARBON_BUDGET_FITTERS[fitter](
        sample_prices,
        sample_emissions,
        threshold,  # type: ignore
    )

    # Cap price between 0.0 and price_too_high_threshold
    if refine_price:
        new_price = min(new_price, price_too_high_threshold)
    new_price = max(new_price, 0.0)

    return round(new_price, resolution)


def linear_fun(x, a, b):
    return a + b * x


def exponential_fun(x, a, b, c):
    return a * np.exp(-b * x) + c


def create_sample(carbon_price, current_emissions, budget, size=4):
    """Calculates a sample of carbon prices to estimate the adjusted carbon price.

    For each of these prices, the equilibrium loop will be run, obtaining a new value
    for the emissions. Out of those price-emissions pairs, the final carbon price will
    be estimated.

    Arguments:
        carbon_price: Current carbon price,
        current_emissions: Current emissions,
        budget: Carbon budget,
        size: Number of points in the sample.

    Returns:
        An array with the sample prices.
    """
    exponent = (current_emissions - budget) / budget
    magnitude = max(1 - np.exp(-exponent), -0.1)
    sample = carbon_price * (1 + np.linspace(0, 1, size) * magnitude)
    return np.abs(sample)


@register_carbon_budget_fitter
def linear(prices: np.ndarray, emissions: np.ndarray, budget: int) -> float:
    """Fits the prices-emissions pairs to a linear function.

    Once that is done, an optimal carbon price is estimated

    Arguments:
        prices: An array with the sample carbon prices,
        emissions: An array with the corresponding emissions,
        budget: The carbon budget for the time period.

    Returns:
        The optimal carbon price.
    """
    guess, weights = linear_guess_and_weights(prices, emissions, budget)

    sol, _ = curve_fit(
        linear_fun, emissions, prices, guess, sigma=weights, absolute_sigma=True
    )

    new_price = linear_fun(budget, *sol)

    return new_price


def linear_guess_and_weights(
    prices: np.ndarray, emissions: np.ndarray, budget: int
) -> tuple:
    """Estimates initial values for the linear fitting algorithm and the weights.

    The points closest to the budget are used to estimate the initial guess. They also
    have the highest weight.

    Arguments:
        prices: An array with the sample carbon prices,
        emissions: An array with the corresponding emissions,
        budget: The carbon budget for the time period.

    Returns:
        The initial guess and weights
    """
    weights = np.abs(emissions - budget)
    idx = np.argsort(weights)

    p = prices[idx][:2]
    e = emissions[idx][:2]

    den = e[1] - e[0]
    num = p[1] - p[0]

    if den != 0:
        b = num / den
        a = p[0] - b * e[0]
    else:
        b = 0
        a = p[0]

    return (a, b), weights


@register_carbon_budget_fitter
def exponential(prices: np.ndarray, emissions: np.ndarray, budget: int) -> float:
    """Fits the prices-emissions pairs to an exponential function.

    Once that is done, an optimal carbon price is estimated

    Arguments:
        prices: An array with the sample carbon prices,
        emissions: An array with the corresponding emissions,
        budget: The carbon budget for the time period.

    Returns:
        The optimal carbon price.
    """
    guess, weights = exp_guess_and_weights(prices, emissions, budget)

    sol, _ = curve_fit(
        exponential_fun, emissions, prices, guess, sigma=weights, absolute_sigma=True
    )

    new_price = exponential_fun(budget, *sol)

    return new_price


def exp_guess_and_weights(
    prices: np.ndarray, emissions: np.ndarray, budget: int
) -> tuple:
    """Estimates initial values for the exponential fitting algorithm and the weights.

    The points closest to the budget are used to estimate the initial guess. They also
    have the highest weight.

    Arguments:
        prices: An array with the sample carbon prices,
        emissions: An array with the corresponding emissions,
        budget: The carbon budget for the time period.

    Returns:
        The initial guess and weights
    """
    weights = np.abs(emissions - budget)

    idx = np.argsort(weights)

    p = prices[idx][:2]
    e = emissions[idx][:2]

    b = 1 / e[0]
    e = np.exp(-b * e)

    den = e[1] - e[0]
    num = p[1] - p[0]

    if den != 0:
        a = num / den
        c = p[0] - a * e[0]
    else:
        a = 0
        c = p[0]
    return (a, b, c), weights


@register_carbon_budget_method
def bisection(
    market: xr.Dataset,
    equilibrium: Callable[[xr.Dataset], FindEquilibriumResults],
    carbon_budget: xr.DataArray,
    commodities: list,
    refine_price: bool = False,
    price_too_high_threshold: float = 10,
    max_iterations: int = 5,
    tolerance: float = 0.1,
    early_termination_count: int = 5,
    resolution: int = 2,
    price_penalty: float = 0.1,
) -> float:
    """Applies bisection algorithm to escalate carbon price and meet the budget.

    A carbon market is meant as a pool of emissions for all
    the modelled regions; therefore, the carbon price applies to all modelled regions.
    Bisection applies an iterative estimations of the emissions
    varying the carbon price until convergence or stop criteria
    are met.
    Builds on 'register_carbon_budget_method'.

    Arguments:
        market: Market, with the prices, supply, consumption and demand
        equilibrium: Method for searching market equilibrium
        carbon_budget: DataArray with the carbon budget
        commodities: List of carbon-related commodities
        refine_price: Boolean to decide on whether carbon price should be capped, with
            the upper bound given by price_too_high_threshold
        price_too_high_threshold: Upper limit for carbon price
        max_iterations: Maximum number of iterations for bisection
        tolerance: Maximum permitted deviation of emissions from the budget
        early_termination_count: Will terminate the loop early if the last n solutions
            are the same
        resolution: Number of decimal places to solve the carbon price to
        price_penalty: Penalty factor applied to carbon price when selecting optimal
            solution when convergence isn't reached. Higher values favor lower prices
            when emissions are similar.

    Returns:
        New value of global carbon price
    """
    # Create cache for emissions at different price points
    emissions_cache = EmissionsCache(market, equilibrium, commodities)

    # Carbon price and emissions threshold in the investment year
    future = market.year[-1]
    target = carbon_budget.sel(year=future).values.item()
    price = market.prices.sel(year=future, commodity=commodities).mean().values.item()

    # Test if emissions are already below the budget without imposing a carbon price
    if emissions_cache[0.0] < target:
        message = (
            f"Emissions for the year {int(future)} are already below the carbon budget "
            "without imposing a carbon price. The carbon price has been set to zero."
        )
        getLogger(__name__).warning(message)
        return 0.0

    # Initial lower and upper bounds on carbon price for the bisection algorithm
    lb_price = 0.0
    epsilon = 10**-resolution  # smallest nonzero price
    ub_price = round(max(price, epsilon), resolution)  # i.e. current price

    # Bisection loop
    for _ in range(max_iterations):  # maximum number of iterations before terminating
        # Cap prices between 0.0 and price_too_high_threshold
        if refine_price:
            ub_price = min(ub_price, price_too_high_threshold)
        lb_price = max(lb_price, 0.0)

        # Calculate/retrieve carbon emissions at new bounds
        lb_price_emissions = emissions_cache[lb_price]
        ub_price_emissions = emissions_cache[ub_price]

        # Terminate early if many consecutive emissions are the same
        if len(emissions_cache) >= early_termination_count + 2:
            recent_values = list(emissions_cache.values())[-early_termination_count:]
            if all(x == recent_values[0] for x in recent_values):
                break

        # Exit loop if lower or upper bound on emissions is close to threshold
        if abs(lb_price_emissions - target) <= abs(tolerance * target):
            return lb_price
        if abs(ub_price_emissions - target) <= abs(tolerance * target):
            return ub_price

        # Convergence not yet reached -> calculate new bounds
        lb_price, ub_price = adjust_bounds(
            lb_price,
            ub_price,
            emissions_cache,
            target,
            resolution,
        )

    # If convergence isn't reached, select the price with lowest emissions
    # We apply a small penalty to the carbon price to keep prices low if there are
    # solutions with equal (or almost equal) emissions.
    new_price = min(
        emissions_cache.keys(),
        key=lambda price: emissions_cache[price] + (price * price_penalty),
    )

    # Output the new price to the log
    getLogger(__name__).info(
        f"Final carbon price for {future.item()}: {new_price} "
        f"(emissions = {emissions_cache[new_price]})"
    )

    # Raise warning message
    if all(emissions_cache[k] > target for k in emissions_cache):
        message = (
            f"Carbon budget could not be met for the year {int(future)} "
            f"(budget: {target}, emissions: {emissions_cache[new_price]}). "
            "This may be because there are no processes available that can meet the "
            "budget, or because emissions from capacity installed earlier in the time "
            "horizon is preventing the budget from being met. "
            "The CO2 price in this year should be interpreted with caution."
        )
    else:
        message = (
            f"Carbon budget could not be matched for the year {int(future)} to within "
            "the specified tolerance. "
            "This is sometimes unavoidable due to a discontinuous emissions landscape "
            "which can make the budget unreachable, but can sometimes be "
            "fixed by increasing max_iterations, early_termination_count or resolution."
        )
    getLogger(__name__).warning(message)
    return new_price


class EmissionsCache(dict):
    """Cache of emissions at different price points for bisection algorithm.

    If a price is queried that is not in the cache, it calculates the emissions at that
    price using solve_market and stores the result in the cache.
    """

    def __init__(self, market, equilibrium, commodities):
        super().__init__()
        self.market = market
        self.equilibrium = equilibrium
        self.commodities = commodities

    def __missing__(self, price):
        value = solve_market(self.market, self.equilibrium, self.commodities, price)
        self[price] = value
        return value


def adjust_bounds(
    lb_price: float,
    ub_price: float,
    emissions_cache: dict[float, float],
    target: float,
    resolution: int = 2,
) -> tuple[float, float]:
    """Adjust the bounds of the carbon price for the bisection algorithm.

    As emissions can be a discontinuous function of the carbon price, this method is
    used to improve the solution search when discontinuities are met, improving the
    bounds search.

    Arguments:
        lb_price: Value of carbon price at lower bound
        ub_price: Value of carbon price at upper bound
        emissions_cache: Dictionary of emissions at different price points
        target: Carbon budget
        resolution: Number of decimal places to solve the carbon price to

    Returns:
        New lower and upper bounds for the carbon price.
    """
    lb_price_emissions = emissions_cache[lb_price]
    ub_price_emissions = emissions_cache[ub_price]

    if lb_price_emissions < target and ub_price_emissions < target:
        # Emissions too low at both prices -> decrease prices
        method = decrease_bounds

    elif lb_price_emissions > target and ub_price_emissions > target:
        # Emissions too high at both prices -> increase prices
        method = increase_bounds

    elif lb_price_emissions > target and ub_price_emissions < target:
        # Threshold is between bounds -> perform bisection
        method = bisect_bounds

    elif lb_price_emissions < target and ub_price_emissions > target:
        # Inverted bounds (i.e. increasing price leads to increasing emissions)
        # Unlikely case, but included for completeness
        method = bisect_bounds_inverted

    else:
        # lb_price_emissions or ub_price_emissions is equal to the target, so we can
        # return the bounds unchanged
        return lb_price, ub_price

    return method(
        lb_price,
        ub_price,
        emissions_cache,
        target,
        resolution,
    )


def decrease_bounds(
    lb_price: float,
    ub_price: float,
    emissions_cache: dict[float, float],
    target: float,
    resolution: int = 2,
) -> tuple[float, float]:
    """Decreases the lb of the carbon price, and sets the ub to the previous lb."""
    denominator = max(target, 1e-3)
    lb_price_emissions = emissions_cache[lb_price]
    exponent = (lb_price_emissions - target) / abs(denominator)  # will be negative
    exponent = max(exponent, -1)  # cap exponent at -1
    exponent = min(exponent, -0.1)  # upper bound cap to force exploration
    ub_price = lb_price
    lb_price = round(lb_price * np.exp(exponent), resolution)
    if lb_price == ub_price:
        lb_price -= 10**-resolution
    return lb_price, ub_price


def increase_bounds(
    lb_price: float,
    ub_price: float,
    emissions_cache: dict[float, float],
    target: float,
    resolution: int = 2,
) -> tuple[float, float]:
    """Increases the ub of the carbon price, and sets the lb to the previous ub."""
    denominator = max(target, 1e-3)
    ub_price_emissions = emissions_cache[ub_price]
    exponent = (ub_price_emissions - target) / abs(denominator)  # will be positive
    exponent = min(exponent, 1)  # cap exponent at 1
    exponent = max(exponent, 0.1)  # lower bound cap to force exploration
    lb_price = ub_price
    ub_price = round(ub_price * np.exp(exponent), resolution)
    if ub_price == lb_price:
        ub_price += 10**-resolution
    return lb_price, ub_price


def bisect_bounds(
    lb_price: float,
    ub_price: float,
    emissions_cache: dict[float, float],
    target: float,
    resolution: int = 2,
) -> tuple[float, float]:
    """Bisects the bounds of the carbon price."""
    midpoint = round((lb_price + ub_price) / 2.0, resolution)
    midpoint_emissions = emissions_cache[midpoint]
    if midpoint_emissions < target:
        ub_price = midpoint
    else:
        lb_price = midpoint
    return lb_price, ub_price


def bisect_bounds_inverted(
    lb_price: float,
    ub_price: float,
    emissions_cache: dict[float, float],
    target: float,
    resolution: int = 2,
) -> tuple[float, float]:
    """Bisects the bounds of the carbon price, in the case of inverted bounds."""
    midpoint = round((lb_price + ub_price) / 2.0, resolution)
    midpoint_emissions = emissions_cache[midpoint]
    if midpoint_emissions > target:
        ub_price = midpoint
    else:
        lb_price = midpoint
    return lb_price, ub_price


def solve_market(
    market: xr.Dataset,
    equilibrium: Callable[[xr.Dataset], FindEquilibriumResults],
    commodities: list,
    carbon_price: float,
) -> float:
    """Solves the market with a new carbon price and returns the emissions.

    Arguments:
        market: Market, with the prices, supply, consumption and demand
        equilibrium: Method for searching market equilibrium
        commodities: List of carbon-related commodities
        carbon_price: New carbon price

    Returns:
        Emissions at the new carbon price.
    """
    future = market.year[-1]
    new_market = market.copy(deep=True)

    # Assign new carbon price and solve market
    new_market.prices.loc[{"year": future, "commodity": commodities}] = carbon_price
    new_market = equilibrium(new_market).market
    new_emissions = (
        new_market.supply.sel(year=future, commodity=commodities)
        .sum(["region", "timeslice", "commodity"])
        .round(decimals=2)
    ).values.item()

    # Log the emissions
    getLogger(__name__).info(
        f"Emissions in {future.item()} with carbon price {carbon_price}: "
        f"{new_emissions}"
    )

    return new_emissions



"""Collection of functions for calculating cost metrics (e.g. LCOE, EAC).

All costs functions take a subset of the following arguments:
- technologies: xr.Dataset of technology parameters
- prices: xr.DataArray with commodity prices
- capacity: xr.DataArray with the capacity of the technologies
- production: xr.DataArray with commodity production by the technologies
- consumption: xr.DataArray with commodity consumption by the technologies
- method: "lifetime" or "annual"

Data should only be provided for a single year (i.e. no "year" dimension in any of the
inputs). Prices, production and consumption data should be split across timeslices
(i.e. have a "timeslice" dimension). Technology parameters may also be specified at
the timeslice level, but capacity should not be.

The `technologies` input will usually contain data for multiple technologies and have
a "technology" dimension (sometimes called "asset" or "replacement"). In this case,
it's important that the `capacity`, `production`, and `consumption` inputs have a
similar dimension to ensure that costs are calculated for all technologies and to
prevent unwanted broadcasting.

Additional dimensions (such as "region") may be present in the inputs, but it's up to
the parent functions to ensure that these are consistent between inputs to prevent
unwanted broadcasting.

The dimensions of the output will be the sum of all dimensions from the input data,
minus "commodity", plus "timeslice" (if not already present).

Some functions have a `method` argument, which can be either ``"annual"`` or
``"lifetime"``. In brief:

- ``annual``: calculates the cost in a single year.
- ``lifetime``: calculates the total cost over the lifetime of the technology,
  using the `technical_life` attribute from the `technologies` dataset. In this
  case, technology parameters, production, consumption, capacity and prices are
  assumed constant over the lifetime; annual costs are discounted using the
  `interest_rate` attribute from the `technologies` dataset and summed across years.

Capital costs are different, as these are a one-time cost for the lifetime of the
technology. These can be annualized by dividing by `technical_life`.

Some functions can calculate both lifetime and annual costs (use the ``method``
argument to select); others implement only one of these modes (see individual
function docstrings for details).

"""

from __future__ import annotations

from functools import wraps

import numpy as np
import xarray as xr

from muse.commodities import is_enduse, is_fuel, is_material, is_pollutant
from muse.quantities import production_amplitude
from muse.timeslices import broadcast_timeslice, distribute_timeslice, get_level


def cost(func):
    """Decorator to validate the output dimensions of the cost functions."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        result = func(*args, **kwargs)

        # Check dimensions of the result
        assert "year" not in result.dims
        assert "commodity" not in result.dims

        # Check that there are no infs or nans in the result
        assert not result.isnull().any()
        assert not np.isinf(result).any()
        return result

    return wrapper


@cost
def capital_costs(
    technologies: xr.Dataset,
    capacity: xr.DataArray,
    method: str = "lifetime",
):
    """Calculate capital costs for the relevant technologies.

    This is the cost of installing each technology to the level specified by the
    `capacity` input.

    Method can be "lifetime" or "annual":

    - ``lifetime``: returns the full capital costs.
    - ``annual``: total capital costs are multiplied by the capital recovery factor to
      obtain annualized costs.
    """
    if method not in ["lifetime", "annual"]:
        raise ValueError("method must be either 'lifetime' or 'annual'.")

    _capital_costs = technologies.cap_par * (capacity**technologies.cap_exp)
    if method == "annual":
        crf = capital_recovery_factor(technologies)
        _capital_costs = _capital_costs * crf
    assert "timeslice" not in _capital_costs.dims
    return _capital_costs


@cost
def environmental_costs(
    technologies: xr.Dataset, prices: xr.DataArray, production: xr.DataArray
) -> xr.DataArray:
    """Calculate timeslice-level environmental costs for the relevant technologies.

    This is the total production of pollutants (commodities flagged by `is_pollutant`)
    multiplied by their prices.
    """
    environmentals = is_pollutant(technologies.comm_usage)
    prices_environmental = prices.sel(commodity=environmentals)
    result = (production * prices_environmental).sum("commodity")
    assert "timeslice" in result.dims
    return result


@cost
def fuel_costs(
    technologies: xr.Dataset, prices: xr.DataArray, consumption: xr.DataArray
) -> xr.DataArray:
    """Calculate timeslice-level fuel costs for the relevant technologies.

    This is the total consumption of fuels (commodities flagged by `is_fuel`)
    multiplied by their prices.
    """
    fuels = is_fuel(technologies.comm_usage)
    prices_fuel = prices.sel(commodity=fuels)
    result = (consumption * prices_fuel).sum("commodity")
    assert "timeslice" in result.dims
    return result


@cost
def material_costs(
    technologies: xr.Dataset, prices: xr.DataArray, consumption: xr.DataArray
) -> xr.DataArray:
    """Calculate timeslice-level material costs for the relevant technologies.

    This is the total consumption of materials (commodities flagged by `is_material`)
    multiplied by their prices.
    """
    material = is_material(technologies.comm_usage)
    prices_material = prices.sel(commodity=material)
    result = (consumption * prices_material).sum("commodity")
    assert "timeslice" in result.dims
    return result


@cost
def fixed_costs(technologies: xr.Dataset, capacity: xr.DataArray) -> xr.DataArray:
    """Calculate annual fixed costs for the relevant technologies.

    This is the fixed running cost over the course of a year corresponding to the
    `fix_par` and `fix_exp` technology parameters.
    """
    result = technologies.fix_par * (capacity**technologies.fix_exp)
    assert "timeslice" not in result.dims
    return result


@cost
def variable_costs(
    technologies: xr.Dataset,
    production: xr.DataArray,
) -> xr.DataArray:
    """Calculate annual variable costs for the relevant technologies.

    This is the cost associated with the `var_par` and `var_exp` technology
    parameters.

    The `production_amplitude` function is first used to calculate technology activity
    based on `production`. This is then used to scale the variable costs.
    """
    tech_activity = production_amplitude(production, technologies).sum("timeslice")
    result = technologies.var_par * tech_activity**technologies.var_exp
    assert "timeslice" not in result.dims
    return result


@cost
def running_costs(
    technologies: xr.Dataset,
    prices: xr.DataArray,
    capacity: xr.DataArray,
    production: xr.DataArray,
    consumption: xr.DataArray,
    aggregate_timeslices: bool = False,
) -> xr.DataArray:
    """Total annual running costs (excluding capital costs).

    This is the sum of environmental, fuel, material, fixed and variable costs.

    .. seealso::
        :py:func:`environmental_costs`
        :py:func:`fuel_costs`
        :py:func:`material_costs`
        :py:func:`fixed_costs`
        :py:func:`variable_costs`
    """
    # Costs associated with commodity inputs and outputs (timeslice-level)
    _environmental_costs = environmental_costs(technologies, prices, production)
    _fuel_costs = fuel_costs(technologies, prices, consumption)
    _material_costs = material_costs(technologies, prices, consumption)

    # Aggregate over timeslices (if required)
    if aggregate_timeslices:
        _environmental_costs = _environmental_costs.sum("timeslice")
        _fuel_costs = _fuel_costs.sum("timeslice")
        _material_costs = _material_costs.sum("timeslice")

    # Costs associated with capacity and production level (annual)
    _fixed_costs = fixed_costs(technologies, capacity)
    _variable_costs = variable_costs(technologies, production)

    # Split fixed/variable across timeslices in proportion to production (if required)
    if not aggregate_timeslices:
        timeslice_level = get_level(production)
        tech_activity = production_amplitude(production, technologies)
        _fixed_costs = distribute_timeslice(
            _fixed_costs, ts=tech_activity, level=timeslice_level
        )
        _variable_costs = distribute_timeslice(
            _variable_costs, ts=tech_activity, level=timeslice_level
        )

    # Total running costs
    result = (
        _environmental_costs
        + _fuel_costs
        + _material_costs
        + _fixed_costs
        + _variable_costs
    )
    return result


@cost
def net_present_value(
    technologies: xr.Dataset,
    prices: xr.DataArray,
    capacity: xr.DataArray,
    production: xr.DataArray,
    consumption: xr.DataArray,
    aggregate_timeslices: bool = False,
) -> xr.DataArray:
    """Net present value (NPV) of the relevant technologies.

     The net present value of a technology is the present value of all the revenues that
     a technology earns over its lifetime minus all the costs of installing and
     operating it. Follows the definition of the `net present cost`_ given by HOMER
     Energy.

     .. _net present cost:
         https://www.homerenergy.com/products/pro/docs/3.15/net_present_cost.html

    - energy commodities INPUTS are related to fuel costs
    - environmental commodities OUTPUTS are related to environmental costs
    - material and service commodities INPUTS are related to consumable costs
    - fixed and variable costs are given as technodata inputs and depend on the
      installed capacity and production (non-environmental), respectively
    - capacity costs are given as technodata inputs and depend on the installed capacity

    .. seealso::
        :py:func:`capital_costs`
        :py:func:`running_costs`

    Arguments:
        technologies: xr.Dataset of technology parameters
        prices: xr.DataArray with commodity prices
        capacity: xr.DataArray with the capacity of the relevant technologies
        production: xr.DataArray with commodity production by the relevant technologies
        consumption: xr.DataArray with commodity consumption by the relevant
            technologies
        aggregate_timeslices: If True, the LCOE is aggregated over timeslices (result
            will not have a "timeslice" dimension)

    Return:
        xr.DataArray with the NPV calculated for the relevant technologies
    """
    # Capital costs (lifetime)
    _capital_costs = capital_costs(technologies, capacity, method="lifetime")

    # Split capital costs across timeslices in proportion to production (if required)
    if not aggregate_timeslices:
        tech_activity = production_amplitude(production, technologies)
        _capital_costs = distribute_timeslice(
            _capital_costs, ts=tech_activity, level=get_level(production)
        )

    # Revenue (annual)
    products = is_enduse(technologies.comm_usage)
    prices_non_env = prices.sel(commodity=products)
    revenues = (production * prices_non_env).sum("commodity")
    if aggregate_timeslices:
        revenues = revenues.sum("timeslice")

    # Running costs (annual)
    _running_costs = running_costs(
        technologies,
        prices,
        capacity,
        production,
        consumption,
        aggregate_timeslices,
    )

    # Calculate running costs and revenues over lifetime
    _running_costs = annual_to_lifetime(_running_costs, technologies)
    revenues = annual_to_lifetime(revenues, technologies)

    # Net present value
    result = revenues - (_capital_costs + _running_costs)
    return result


@cost
def net_present_cost(
    technologies: xr.Dataset,
    prices: xr.DataArray,
    capacity: xr.DataArray,
    production: xr.DataArray,
    consumption: xr.DataArray,
    aggregate_timeslices: bool = False,
) -> xr.DataArray:
    """Net present cost (NPC) of the relevant technologies.

    The net present cost of a Component is the present value of all the costs of
    installing and operating the Component over the project lifetime, minus the present
    value of all the revenues that it earns over the project lifetime.

    .. seealso::
        :py:func:`net_present_value`.

    Arguments:
        technologies: xr.Dataset of technology parameters
        prices: xr.DataArray with commodity prices
        capacity: xr.DataArray with the capacity of the relevant technologies
        production: xr.DataArray with commodity production by the relevant technologies
        consumption: xr.DataArray with commodity consumption by the relevant
            technologies
        aggregate_timeslices: If True, the LCOE is aggregated over timeslices (result
            will not have a "timeslice" dimension)

    Return:
        xr.DataArray with the NPC calculated for the relevant technologies
    """
    result = -net_present_value(
        technologies,
        prices,
        capacity,
        production,
        consumption,
        aggregate_timeslices,
    )
    return result


@cost
def equivalent_annual_cost(
    technologies: xr.Dataset,
    prices: xr.DataArray,
    capacity: xr.DataArray,
    production: xr.DataArray,
    consumption: xr.DataArray,
    aggregate_timeslices: bool = False,
) -> xr.DataArray:
    """Equivalent annual costs (or annualized cost) of a technology.

    This is the cost that, if it were to occur equally in every year of the
    project lifetime, would give the same net present cost as the actual cash
    flow sequence associated with that component. The cost is computed using the
    `annualized cost`_ expression given by HOMER Energy.

    .. _annualized cost:
        https://www.homerenergy.com/products/pro/docs/3.15/annualized_cost.html

    .. seealso::
        :py:func:`net_present_cost`

    Arguments:
        technologies: xr.Dataset of technology parameters
        prices: xr.DataArray with commodity prices
        capacity: xr.DataArray with the capacity of the relevant technologies
        production: xr.DataArray with commodity production by the relevant technologies
        consumption: xr.DataArray with commodity consumption by the relevant
            technologies
        aggregate_timeslices: If True, the LCOE is aggregated over timeslices (result
            will not have a "timeslice" dimension)

    Return:
        xr.DataArray with the EAC calculated for the relevant technologies
    """
    npc = net_present_cost(
        technologies,
        prices,
        capacity,
        production,
        consumption,
        aggregate_timeslices,
    )
    crf = capital_recovery_factor(technologies)
    if not aggregate_timeslices:
        crf = broadcast_timeslice(crf, level=get_level(production))
    result = npc * crf
    return result


@cost
def levelized_cost_of_energy(
    technologies: xr.Dataset,
    prices: xr.DataArray,
    capacity: xr.DataArray,
    production: xr.DataArray,
    consumption: xr.DataArray,
    method: str = "lifetime",
    aggregate_timeslices: bool = False,
) -> xr.DataArray:
    """Levelized cost of energy (LCOE) of technologies over their lifetime.

    It follows the `simplified LCOE` given by NREL.

    .. seealso::
        :py:func:`capital_costs`
        :py:func:`running_costs`

    Can calculate either a lifetime or annual LCOE.

    - ``lifetime``: the average cost per unit of production over the entire lifetime
      of the technology. Annual running costs and production are calculated for the
      full lifetime and adjusted to present value using the discount rate. Total
      costs (running costs over the lifetime + initial capital costs) are divided by
      total production to obtain the average cost per unit.

    - ``annual``: the average cost per unit of production in a single year. Annual
      running costs and production are calculated for a single year, capital costs
      are annualized using the capital recovery factor, and total costs are divided
      by production to obtain the average cost per unit.

    Arguments:
        technologies: xr.Dataset of technology parameters
        prices: xr.DataArray with commodity prices
        capacity: xr.DataArray with the capacity of the relevant technologies
        production: xr.DataArray with commodity production by the relevant technologies
        consumption: xr.DataArray with commodity consumption by the relevant
            technologies
        method: "lifetime" or "annual"
        aggregate_timeslices: If True, the LCOE is aggregated over timeslices (result
            will not have a "timeslice" dimension)

    Return:
        xr.DataArray with the LCOE calculated for the relevant technologies
    """
    if method not in ["lifetime", "annual"]:
        raise ValueError("method must be either 'lifetime' or 'annual'.")

    # Capital costs (lifetime or annual depending on method)
    _capital_costs = capital_costs(technologies, capacity, method)

    # Split capital costs across timeslices in proportion to production (if required)
    if not aggregate_timeslices:
        tech_activity = production_amplitude(production, technologies)
        _capital_costs = distribute_timeslice(
            _capital_costs, ts=tech_activity, level=get_level(production)
        )

    # Running costs (annual)
    _running_costs = running_costs(
        technologies, prices, capacity, production, consumption, aggregate_timeslices
    )

    # Production (annual)
    products = is_enduse(technologies.comm_usage)
    prod = (
        production.where(production > 0.0, 1e-6)
        .sel(commodity=products)
        .sum(
            "commodity"
        )  # TODO: is this the correct way to deal with multiple products?
    )
    if aggregate_timeslices:
        prod = prod.sum("timeslice")

    # If method is lifetime, have to adjust running costs and production
    if method == "lifetime":
        _running_costs = annual_to_lifetime(_running_costs, technologies)
        prod = annual_to_lifetime(prod, technologies)

    # LCOE
    result = (_capital_costs + _running_costs) / prod
    return result


def supply_cost(
    production: xr.DataArray, lcoe: xr.DataArray, asset_dim: str | None = "asset"
) -> xr.DataArray:
    """Supply cost given production and the levelized cost of energy.

    In practice, the supply cost is the weighted average LCOE over assets (`asset_dim`),
    where the weights are the production.

    Very low costs are set to zero.

    Arguments:
        production: Amount of goods produced. In practice, production can be obtained
            from the capacity for each asset via the method
            `muse.quantities.production`.
        lcoe: Levelized cost of energy for each good produced. In practice, it can be
            obtained from market prices via
            `muse.costs.levelized_cost_of_energy`.
        asset_dim: Name of the dimension(s) holding assets, processes or technologies.
    """
    data = xr.Dataset(dict(production=production, prices=production * lcoe))
    if asset_dim is not None:
        if "region" not in data.coords or len(data.region.dims) == 0:
            data = data.sum(asset_dim)
        else:
            data = data.groupby("region").sum(asset_dim)

    costs = data.prices / data.production.where(np.abs(data.production) > 1e-15, np.inf)
    return costs.where(costs > 1e-4, 0)


def capital_recovery_factor(technologies: xr.Dataset) -> xr.DataArray:
    """Capital recovery factor using interest rate and expected lifetime.

    The `capital recovery factor`_ is computed using the expression given by HOMER
    Energy.

    .. _capital recovery factor:
        https://www.homerenergy.com/products/pro/docs/3.15/capital_recovery_factor.html

    If the interest rate is zero, this simplifies to 1 / nyears

    Arguments:
        technologies: All the technologies

    Return:
        xr.DataArray with the CRF calculated for the relevant technologies
    """
    nyears = technologies.technical_life.astype(int)
    interest_rate = technologies.interest_rate

    crf = xr.where(
        interest_rate == 0,
        1 / nyears,
        interest_rate / (1 - (1 / (1 + interest_rate) ** nyears)),
    )

    assert "year" not in crf.dims
    return crf


def annual_to_lifetime(costs: xr.DataArray, technologies: xr.Dataset):
    """Convert annual costs to lifetime costs.

    Costs are provided for a single year. These same costs are assumed to apply for the
    full lifetime of the technologies, subject to a discount factor. The costs are then
    summed over the lifetime of the technologies.

    Args:
        costs: xr.DataArray of costs for a single year.
        technologies: xr.Dataset of technology parameters
    """
    assert "year" not in costs.dims
    assert "year" not in technologies.dims
    life = technologies.technical_life.astype(int)
    iyears = range(life.values.max())
    years = xr.DataArray(iyears, coords={"year": iyears}, dims="year")
    rates = discount_factor(
        years=years,
        interest_rate=technologies.interest_rate,
        mask=years <= life,
    )
    if "timeslice" in costs.dims:
        rates = broadcast_timeslice(rates, level=get_level(costs))
    return (costs * rates).sum("year")


def discount_factor(
    years: xr.DataArray, interest_rate: xr.DataArray, mask: xr.DataArray | None = None
):
    """Calculate an array with of discount factor values over the years.

    Args:
        years: xr.DataArray with the years counting from the present year
            (i.e. current year = 0)
        interest_rate: xr.DataArray with the interest rate for different technologies
        mask: Optional mask to apply to the result (e.g. cutting to zero after the
            technology lifetime)
    """
    assert set(years.dims) == {"year"}
    assert "year" not in interest_rate.dims

    # Calculate discount factor over the years
    df = 1 / (1 + interest_rate) ** years

    # Apply mask
    if mask is not None:
        assert set(mask.dims) == set(interest_rate.dims) | {"year"}
        df = df * mask
    return df



class AgentShareNotDefined(Exception):
    """Indicates that an agent share is missing from a technodata file."""

    msg = """All agents must be represented in the technodata file.
If using "New" and "Retrofit" agents, you need a column with the name of each "Retrofit"
agent share. If only using "New" agents, you need a column with the name of each "New"
agent share. Please check that all agents are represented in the technodata file, and
that the agent share names match those specified in your agents file.
"""

    def __str__(self):
        return self.msg


class GrowthOfCapacityTooConstrained(Exception):
    """Indicates that the investment step failed because capacity could not grow."""

    msg = """Error during the investment process. The capacity was not allowed to grow
sufficiently in order to match the demand. Consider increasing the MaxCapacityAddition
and/or the MaxCapacityGrowth in the technodata."""

    def __str__(self):
        return self.msg


class TechnologyNotDefined(Exception):
    """Indicates that the initialisation fails because a technology is not found."""

    msg = """Error during the initialisation of a sector.
The model tries to assign a share of the total capacity to an agent but it
cannot find a match between technodata and existing capacity.
Check the spelling of the technology names."""

    def __str__(self):
        return self.msg


class FailedInterpolation(Exception):
    """Indicates that the initialisation fails due to interpolation."""

    msg = """Error during the initialisation of a sector.
The model tries to interpolate values in time of the technologies.
It fails to interpolate some parameters for selected years.
This results in nans in the datasets of technology data.
Check the spelling of the technology names in the sector data."""

    def __str__(self):
        return self.msg


class RetrofitAgentInStandardDemandShare(Exception):
    msg = """A retrofit agent has been found in a 'New agents'-only demand share
function. Make sure you remove all the retro agents from the Agents input files or use a
demand share method that can handle both new and retro agents."""

    def __str__(self):
        return self.msg


class AgentWithNoAssetsInDemandShare(Exception):
    msg = """This error refers to an agent with no assets. To fix this error, check the
capacity assignment to the agents. One possibility is that you have decided not
to use "Retrofit" agents, as such you may have already removed them from the
agent definition file and the file of technodata, the system TOML file should
change the demand_share to "standard_demand" function in each subsector
section for each of the selected sectors to model."""

    def __str__(self):
        return self.msg


class NoInteractionsFound(Exception):
    msg = """A network with no interactions has been found. This might be the case if
there are no retrofit agents and yet a 'new_to_retro' network has been defined for a
particular sector. Asses the existence of both new and retrofit agents for all sectors
and remove the new_to_retro interacton network if it is not needed """

    def __str__(self):
        return self.msg



def test_from_technologies(technologies, coords):
    """Test CommodityUsage.from_technologies method.

    Verifies that commodity usage flags are correctly set based on technology
    inputs/outputs and commodity types.
    """
    from muse.commodities import CommodityUsage

    technologies = technologies.drop_vars("comm_usage")
    technologies["commodity_type"] = "commodity", coords["commodity_type"]
    technologies = technologies.set_coords("commodity_type")

    comm_usage = CommodityUsage.from_technologies(technologies)

    # Check which commodities are used in any region/year/technology
    usage_mask = (
        technologies[["fixed_inputs", "fixed_outputs", "flexible_inputs"]] > 0
    ).any(("region", "year", "technology"))

    # Test individual commodity usage flags
    for actual, is_consumable, is_product in zip(
        comm_usage,
        usage_mask.fixed_inputs | usage_mask.flexible_inputs,
        usage_mask.fixed_outputs,
    ):
        assert bool(actual & CommodityUsage.PRODUCT) == is_product
        assert bool(actual & CommodityUsage.CONSUMABLE) == is_consumable

    # Test commodity type flags across all items
    for i in range(len(comm_usage)):
        assert (
            bool(comm_usage[i] & CommodityUsage.PRODUCT) == usage_mask.fixed_outputs[i]
        )
        assert bool(comm_usage[i] & CommodityUsage.ENVIRONMENTAL) == (
            usage_mask.commodity_type[i] == "environmental"
        )
        assert bool(comm_usage[i] & CommodityUsage.ENERGY) == (
            usage_mask.commodity_type[i] == "energy"
        )



import xarray as xr
from numpy import arange
from numpy.random import choice, randint, random
from pytest import approx, fixture

from muse.demand_matching import demand_matching


def assert_matches_demand(result, demand):
    """Helper function to check if result matches demand after broadcasting."""
    actual, expected = xr.broadcast(result, demand)
    assert actual.data == approx(expected.data)


@fixture(params=["without c", "with c"])
def demand(request):
    demand = xr.DataArray(
        randint(0, 10, (5, 3, 4)),
        coords={
            "a": choice(range(15), 5, replace=False),
            "b": choice(list("abcdefghi"), 3, replace=False),
            "c": choice(["🐮", "🐄", "🕙", "📕", "💤", "🐜", "🐘"], 4, replace=False),
        },
        dims=("a", "b", "c"),
    )
    return demand.sum("c") if request.param == "without c" else demand.astype("float64")


@fixture
def cost(demand):
    return xr.DataArray(
        randint(0, 10, size=demand.shape), coords=demand.coords, dims=demand.dims
    )


@fixture
def dataset(timeslice):
    """Creates a test dataset with cost, demand, and max_production."""
    ds = xr.Dataset()

    # Create dimension coordinates
    for dim in ("d", "m", "c", "dc", "dm", "cm", "dcm"):
        nitems = randint(1, 10)
        ds[dim] = dim, choice(range(2 * nitems), nitems, replace=False)

    # Set up cost array
    shape_cost = (len(ds.c), len(ds.dc), len(ds.cm), len(ds.dcm))
    ds["cost"] = (("c", "dc", "cm", "dcm"), randint(0, 5, shape_cost).astype("float64"))

    # Set up max_production with some zero values
    shape_prod = (len(ds.m), len(ds.dm), len(ds.cm), len(ds.dcm))
    ds["max_production"] = (
        ("m", "dm", "cm", "dcm"),
        1.0 * randint(0, 10, shape_prod) * (0 == randint(0, 2, shape_prod)),
    )

    # Calculate demand based on production
    summed_production = ds.max_production.sum(("m", "cm"))

    # Create shares with correct shapes
    dc_share = xr.DataArray(random(len(ds.dc)), coords={"dc": ds.dc}, dims="dc")
    dc_share = dc_share / dc_share.sum()

    d_share = xr.DataArray(random(len(ds.d)), coords={"d": ds.d}, dims="d")
    d_share = d_share / d_share.sum()

    # Broadcast and multiply
    ds["demand"] = summed_production * dc_share * d_share

    return ds


def test_cost_order(dataset):
    """Tests the ordering of technology selection based on cost.

    This test verifies that:
    1. When costs are ordered, the lowest cost technology is selected first
    2. When two technologies have equal costs, they share the demand equally
    3. When a technology becomes more expensive, it is not selected
    """
    # Simplify dataset to focus on technology selection dimensions
    ds = dataset.sum(set(dataset.dims).difference(("dm", "cm")))
    if ds.cm.size < 2:
        return  # Skip test if insufficient technologies for comparison

    # Set up initial conditions where any tech can fulfill total demand
    total_demand = ds.demand.sum()
    ds.max_production[:] = total_demand

    # Test Case 1: Sequential cost ordering - should select first tech
    ds.cost[:] = arange(ds.cost.size).reshape(ds.cost.shape)
    result = demand_matching(ds.demand, ds.cost)
    assert_matches_demand(result.sum(ds.cost.dims), ds.demand)
    assert result.sum(ds.demand.dims)[0].data == approx(total_demand.data)

    # Test Case 2: Equal costs for first two techs - should split demand equally
    ds.cost[0] = 1
    result = demand_matching(ds.demand, ds.cost)
    assert_matches_demand(result.sum(ds.cost.dims), ds.demand)
    summed = result.sum(ds.demand.dims).data
    assert summed[0] == approx(summed[1]), (
        "First two technologies should share demand equally"
    )
    assert 2 * summed[0] == approx(total_demand.data), (
        "Sum of shared demand should equal total"
    )

    # Test Case 3: First tech more expensive - should select second tech only
    ds.cost[0] = 2
    result = demand_matching(ds.demand, ds.cost)
    assert_matches_demand(result.sum(ds.cost.dims), ds.demand)
    summed = result.sum(ds.demand.dims).data
    assert summed[1] == approx(total_demand.data), (
        "Second technology should fulfill all demand"
    )


def test_no_constraints_no_i_dims(demand, cost):
    # Test without b dimension in cost
    result = demand_matching(demand, cost.sum("b"))
    assert set(result.dims) == set(demand.dims)
    assert_matches_demand(result, demand)

    # Test with all dimensions
    result = demand_matching(demand, cost)
    assert set(result.dims) == set(demand.dims)
    assert_matches_demand(result, demand)

    # Test with summed a dimension
    result = demand_matching(demand.sum("a"), cost)
    assert set(result.dims) == set(demand.dims)
    assert_matches_demand(result.sum("a"), demand.sum("a"))


def test_no_constraints_i_dims(demand, cost):
    # Test protected dimension 'a'
    result = demand_matching(demand.sum("a"), cost, protected_dims={"a"})
    assert set(result.dims) == set(demand.dims)
    assert_matches_demand(result.sum("a"), demand.sum("a"))

    # Test protected dimension 'b'
    result = demand_matching(demand.sum("b"), cost, protected_dims={"b"})
    assert set(result.dims) == set(demand.dims)
    assert_matches_demand(result.sum("b"), demand.sum("b"))


def test_one_non_binding_constraint(demand, cost):
    """Tests constraints where excess is always 0."""
    # Test with full dimensions
    result = demand_matching(demand, cost, 2 * demand)
    assert set(result.dims) == set(demand.dims)
    assert_matches_demand(result, demand)

    # Test with summed dimension
    result = demand_matching(demand, cost, (2 * demand).sum("a"))
    assert set(result.dims) == set(demand.dims)
    assert_matches_demand(result, demand)


def test_one_cutting_constraint(demand, cost):
    """Tests constraints where excess is not always 0."""
    constraint = demand.sum("a") * 0.5

    # Test with full demand
    result = demand_matching(demand, cost, constraint)
    assert set(result.dims) == set(demand.dims)
    assert (result.sum("a") - constraint <= 1e-12).all()
    expected, actual = xr.broadcast(result.sum("a") + constraint, demand.sum("a"))
    assert actual.values == approx(expected.values)

    # Test with summed demand
    result = demand_matching(demand.sum("b"), cost, constraint)
    assert set(result.dims) == set(demand.dims)
    assert (result.sum("b") - demand.sum("b") < 1e-12).all()


def test_two_cutting_constraint(demand, cost):
    """Tests multiple constraints where excess is not always 0."""
    constraint0 = demand.sum("a") * 0.75
    constraint1 = demand.sum("b") * 0.85 + demand.sum("a") * 0.80

    # Test with full demand
    result = demand_matching(demand, cost, constraint0, constraint1)
    assert set(result.dims) == set(demand.dims)
    assert (result.sum("a") - constraint0 <= 1e-12).all()
    assert (result - constraint1 <= 1e-12).all()
    assert (result - demand <= 1e-12).all()

    # Test with summed demand
    result = demand_matching(demand.sum("a"), cost, constraint0, constraint1)
    assert set(result.dims) == set(demand.dims)
    assert (result.sum("a") - constraint0 <= 1e-12).all()
    assert (result - constraint1 <= 1e-12).all()
    assert (result.sum("a") - demand.sum("a") <= 1e-12).all()



from typing import Union

import numpy as np
import pandas as pd
import xarray as xr
from pytest import approx, fixture, raises

from muse import examples
from muse.constraints import (
    demand,
    demand_limiting_capacity,
    max_capacity_expansion,
    max_production,
    minimum_service,
)
from muse.lp_adapter import (
    ScipyAdapter,
    back_to_muse_quantity,
    lp_constraint_matrix,
    selected_quantity,
    unified_dataset,
)
from muse.timeslices import drop_timeslice
from muse.utilities import broadcast_over_assets, interpolate_capacity, reduce_assets

CURRENT_YEAR = 2020
INVESTMENT_YEAR = 2025


@fixture
def model_data():
    """Model data required for the constraints.

    Returns:
        dict: Contains all necessary data for building constraints:
            - technologies: Technology data for a single year
            - capacity: Capacity data for assets in the current and investment year
            - demand: Demand for the investment year
            - search_space: Search space for the assets
    """
    from muse.quantities import maximum_production

    # Load residential sector data
    residential = examples.sector("residential", model="medium")

    # Extract technologies and assets
    technologies = residential.technologies.squeeze("region").sel(year=INVESTMENT_YEAR)
    assets = next(a.assets for a in residential.agents)

    # Add nonzero minimum service factor data to allow calculation of the constraint
    technologies["minimum_service_factor"] = 0.1 * xr.ones_like(
        technologies.technology, dtype=float
    )

    # Calculate capacity
    capacity = interpolate_capacity(
        reduce_assets(assets.capacity, coords=("technology", "region")),
        year=[CURRENT_YEAR, INVESTMENT_YEAR],
    )

    # Create initial market demand as 80% of maximum production
    market_demand = 0.8 * maximum_production(
        broadcast_over_assets(technologies, assets),
        assets.capacity,
    ).sel(year=INVESTMENT_YEAR).groupby("technology").sum("asset").rename(
        technology="asset"
    )

    # Remove un-demanded commodities
    market_demand = market_demand.sel(
        commodity=(market_demand > 0).any(dim=["timeslice", "asset"])
    )

    # Create search space
    search_space = examples.search_space("residential", "medium")
    search_space = search_space.sel(asset=assets.technology.values)

    # Return dictionary of data
    return {
        "technologies": technologies,
        "capacity": capacity,
        "demand": market_demand,
        "search_space": search_space,
    }


@fixture(params=["timeslice_as_list", "timeslice_as_multindex"])
def constraints(request, model_data):
    """Default set of constraints for testing."""
    constraints = {
        "max_production": max_production(**model_data),
        "demand": demand(**model_data),
        "max_capacity_expansion": max_capacity_expansion(**model_data),
        "demand_limiting_capacity": demand_limiting_capacity(**model_data),
        "minimum_service": minimum_service(**model_data),
    }

    # Testing two different ways of handling timeslices
    if request.param == "timeslice_as_multindex":
        constraints = {key: _as_list(cs) for key, cs in constraints.items()}
    return constraints


def _as_list(data: Union[xr.DataArray, xr.Dataset]) -> Union[xr.DataArray, xr.Dataset]:
    """Helper function to convert timeslice data to list format."""
    if "timeslice" in data.dims:
        data = data.copy(deep=False)
        index = pd.MultiIndex.from_tuples(
            data.get_index("timeslice"), names=("month", "day", "hour")
        )
        mindex_coords = xr.Coordinates.from_pandas_multiindex(index, "timeslice")
        data = drop_timeslice(data).assign_coords(mindex_coords)
    return data


def test_model_data(model_data):
    """Validating that the model data has appropriate dimensions."""
    assert set(model_data["technologies"].dims) == {"technology", "commodity"}
    assert set(model_data["search_space"].dims) == {"asset", "replacement"}
    assert set(model_data["capacity"].dims) == {"asset", "year"}
    assert set(model_data["demand"].dims) == {
        "asset",
        "commodity",
        "timeslice",
    }


def test_constraints_dimensions(constraints):
    """Test dimensions of all constraint matrices."""
    # Max production constraint
    max_prod_dims = {"asset", "commodity", "replacement", "timeslice"}
    assert set(constraints["max_production"].capacity.dims) == max_prod_dims
    assert set(constraints["max_production"].production.dims) == max_prod_dims
    assert set(constraints["max_production"].b.dims) == max_prod_dims

    # Demand constraint
    assert set(constraints["demand"].capacity.dims) == set()
    assert set(constraints["demand"].production.dims) == set()
    assert set(constraints["demand"].b.dims) == {"asset", "commodity", "timeslice"}

    # Demand limiting capacity constraint
    assert set(constraints["demand_limiting_capacity"].capacity.dims) == {
        "asset",
        "commodity",
        "replacement",
    }
    assert set(constraints["demand_limiting_capacity"].production.dims) == set()
    assert set(constraints["demand_limiting_capacity"].b.dims) == {"asset", "commodity"}

    # Max capacity expansion constraint
    assert set(constraints["max_capacity_expansion"].capacity.dims) == set()
    assert set(constraints["max_capacity_expansion"].production.dims) == set()
    assert set(constraints["max_capacity_expansion"].b.dims) == {"replacement"}

    # Minimum service constraint
    assert set(constraints["minimum_service"].capacity.dims) == max_prod_dims
    assert set(constraints["minimum_service"].production.dims) == max_prod_dims
    assert set(constraints["minimum_service"].b.dims) == max_prod_dims


def test_max_capacity_expansion(constraints):
    """Checking basic properties of the max capacity expansion constraint."""
    max_capacity_expansion = constraints["max_capacity_expansion"]
    assert (max_capacity_expansion.capacity == 1).all()
    assert max_capacity_expansion.production == 0
    assert max_capacity_expansion.b.dims == ("replacement",)
    assert max_capacity_expansion.b.shape == (4,)
    assert (
        max_capacity_expansion.replacement
        == ["estove", "gasboiler", "gasstove", "heatpump"]
    ).all()


def test_max_production(constraints):
    """Checking basic properties of the max production constraint."""
    assert (constraints["max_production"].capacity <= 0).all()


def test_demand_limiting_capacity(constraints):
    """Checking basic properties of the demand limiting capacity constraint."""
    demand_limiting_capacity = constraints["demand_limiting_capacity"]
    max_production = constraints["max_production"]
    demand_constraint = constraints["demand"]

    # Test capacity values
    expected_capacity = (
        -max_production.capacity.max("timeslice")
        if "timeslice" in max_production.capacity.dims
        else -max_production.capacity
    )
    assert abs(demand_limiting_capacity.capacity - expected_capacity).sum() < 1e-5

    # Test production and b values
    assert demand_limiting_capacity.production == 0
    expected_b = (
        demand_constraint.b.max("timeslice")
        if "timeslice" in demand_constraint.b.dims
        else demand_constraint.b
    )
    assert abs(demand_limiting_capacity.b - expected_b).sum() < 1e-5


def test_max_capacity_expansion_no_limits(model_data):
    """Checking that the constraint is None when no limits are set."""
    technologies = model_data["technologies"].drop_vars(
        ["max_capacity_addition", "max_capacity_growth", "total_capacity_limit"]
    )
    assert (
        max_capacity_expansion(**{**model_data, "technologies": technologies}) is None
    )


def test_max_capacity_expansion_infinite_limits(model_data):
    """Checking that error is raised when infinite limits are set."""
    technologies = model_data["technologies"].copy()
    for limit in [
        "max_capacity_addition",
        "max_capacity_growth",
        "total_capacity_limit",
    ]:
        technologies[limit] = np.inf
    with raises(ValueError):
        max_capacity_expansion(**{**model_data, "technologies": technologies})


def test_max_capacity_expansion_seed(model_data):
    """Sanity checks for the seed parameter of the max capacity expansion constraint."""
    seed = 10
    technologies = model_data["technologies"].copy()
    technologies["growth_seed"] = seed

    # Test different capacity scenarios
    scenarios = [0, seed, 2 * seed]
    results = []
    for cap in scenarios:
        capacity = model_data["capacity"].copy()
        capacity.sel(year=2020)[:] = cap
        results.append(
            max_capacity_expansion(
                **{
                    **model_data,
                    "technologies": technologies,
                    "capacity": capacity,
                }
            )
        )

    # Zero capacity should match seed capacity
    assert results[0].b.values == approx(results[1].b.values)
    # Higher capacity should differ
    assert results[0].b.values != approx(results[2].b.values)


def test_no_minimum_service(model_data):
    """Checking that the constraint is None when minimum service factor is set to 0."""
    technologies = model_data["technologies"].copy()
    technologies["minimum_service_factor"] = 0
    assert minimum_service(**{**model_data, "technologies": technologies}) is None


@fixture
def lp_inputs(model_data):
    """Inputs to the lp adapter, in addition to the constraints."""
    # Make up capacity costs data
    shape = model_data["search_space"].shape
    costs = model_data["search_space"] * np.arange(np.prod(shape)).reshape(shape)

    # List of commodities
    commodities = list(model_data["demand"].commodity.values)

    return {
        "capacity_costs": costs,
        "commodities": commodities,
    }


@fixture
def lpcosts(lp_inputs):
    """Benchmark lpcosts dataset to test against."""
    from muse.lp_adapter import lp_costs

    return lp_costs(**lp_inputs)


def test_lp_constraints_matrix_b_is_scalar(lpcosts, constraints):
    """B is a scalar - output should be equivalent to a single row matrix."""
    constraint = constraints["max_production"]

    for attr in ["capacity", "production"]:
        lpconstraint = lp_constraint_matrix(
            xr.DataArray(1), getattr(constraint, attr), getattr(lpcosts, attr)
        )
        expected_value = -1 if attr == "capacity" else 1
        assert lpconstraint.values == approx(expected_value)
        assert set(lpconstraint.dims) == {
            f"d({x})" for x in getattr(lpcosts, attr).dims
        }


def test_max_production_constraint_diagonal(lpcosts, constraints):
    """Test production side of max capacity production is diagonal."""
    constraint = constraints["max_production"]

    # Test capacity constraints
    result = lp_constraint_matrix(constraint.b, constraint.capacity, lpcosts.capacity)
    decision_dims = {f"d({x})" for x in lpcosts.capacity.dims}
    constraint_dims = {
        f"c({x})" for x in set(lpcosts.production.dims).union(constraint.b.dims)
    }
    assert set(result.dims) == decision_dims.union(constraint_dims)

    # Test production constraints
    result = lp_constraint_matrix(
        constraint.b, constraint.production, lpcosts.production
    )
    decision_dims = {f"d({x})" for x in lpcosts.production.dims}
    assert set(result.dims) == decision_dims.union(constraint_dims)

    # Verify diagonal matrix
    result = result.reset_index("d(timeslice)", drop=True).assign_coords(
        {"d(timeslice)": result["d(timeslice)"].values}
    )
    stacked = result.stack(d=sorted(decision_dims), c=sorted(constraint_dims))
    assert stacked.shape[0] == stacked.shape[1]
    assert stacked.values == approx(np.eye(stacked.shape[0]))


def test_lp_constraint(lpcosts, constraints):
    from muse.lp_adapter import lp_constraint

    constraint = constraints["max_production"]

    result = lp_constraint(constraint, lpcosts)
    constraint_dims = {
        f"c({x})" for x in set(lpcosts.production.dims).union(constraint.b.dims)
    }

    # Test capacity constraints
    decision_dims = {f"d({x})" for x in lpcosts.capacity.dims}
    assert set(result.capacity.dims) == decision_dims.union(constraint_dims)

    # Test production constraints
    decision_dims = {f"d({x})" for x in lpcosts.production.dims}
    assert set(result.production.dims) == decision_dims.union(constraint_dims)
    stacked = result.production.stack(
        d=sorted(decision_dims), c=sorted(constraint_dims)
    )
    assert stacked.shape[0] == stacked.shape[1]
    assert stacked.values == approx(np.eye(stacked.shape[0]))

    assert set(result.b.dims) == constraint_dims
    assert result.b.values == approx(0)


def test_to_scipy_adapter_maxprod(lp_inputs, lpcosts, constraints):
    """Test scipy adapter with max production constraint."""
    adapter = ScipyAdapter.from_muse_data(
        **lp_inputs,
        constraints=[constraints["max_production"]],
    )

    assert set(adapter.kwargs) == {"c", "A_ub", "b_ub", "A_eq", "b_eq", "bounds"}
    assert adapter.bounds == (0, np.inf)
    assert adapter.A_ub is not None
    assert adapter.b_ub is not None
    assert adapter.A_eq is None
    assert adapter.b_eq is None
    assert adapter.c.ndim == 1
    assert adapter.b_ub.ndim == 1
    assert adapter.A_ub.ndim == 2

    capsize = lpcosts.capacity.size  # number of capacity decision variables
    prodsize = lpcosts.production.size  # number of production decision variables
    bsize = adapter.b_ub.size  # number of constraints

    assert adapter.c.size == capsize + prodsize
    assert adapter.A_ub.shape[0] == bsize
    assert adapter.A_ub.shape[1] == capsize + prodsize

    assert (
        bsize
        == lpcosts.commodity.size
        * lpcosts.timeslice.size
        * lpcosts.asset.size
        * lpcosts.replacement.size
    )
    assert adapter.b_ub == approx(0)
    assert adapter.A_ub[:, capsize:] == approx(np.eye(prodsize))


def test_to_scipy_adapter_demand(lp_inputs, lpcosts, constraints):
    """Test scipy adapter with demand constraint."""
    adapter = ScipyAdapter.from_muse_data(
        **lp_inputs,
        constraints=[constraints["demand"]],
    )

    assert set(adapter.kwargs) == {"c", "A_ub", "b_ub", "A_eq", "b_eq", "bounds"}
    assert adapter.bounds == (0, np.inf)
    assert adapter.A_ub is not None
    assert adapter.b_ub is not None
    assert adapter.A_eq is None
    assert adapter.b_eq is None
    assert adapter.c.ndim == 1
    assert adapter.b_ub.ndim == 1
    assert adapter.A_ub.ndim == 2

    capsize = lpcosts.capacity.size  # number of capacity decision variables
    prodsize = lpcosts.production.size  # number of production decision variables
    bsize = adapter.b_ub.size  # number of constraints

    assert adapter.c.size == capsize + prodsize
    assert adapter.A_ub.shape[0] == bsize
    assert adapter.A_ub.shape[1] == capsize + prodsize

    assert bsize == lpcosts.commodity.size * lpcosts.timeslice.size * lpcosts.asset.size
    assert adapter.A_ub[:, :capsize] == approx(0)
    assert set(adapter.A_ub[:, capsize:].flatten()) == {0.0, -1.0}


def test_to_scipy_adapter_max_capacity_expansion(lp_inputs, lpcosts, constraints):
    """Test scipy adapter with max capacity expansion constraint."""
    adapter = ScipyAdapter.from_muse_data(
        **lp_inputs,
        constraints=[constraints["max_capacity_expansion"]],
    )

    assert set(adapter.kwargs) == {"c", "A_ub", "b_ub", "A_eq", "b_eq", "bounds"}
    assert adapter.bounds == (0, np.inf)
    assert adapter.A_ub is not None
    assert adapter.b_ub is not None
    assert adapter.A_eq is None
    assert adapter.b_eq is None
    assert adapter.c.ndim == 1
    assert adapter.b_ub.ndim == 1
    assert adapter.A_ub.ndim == 2

    capsize = lpcosts.capacity.size  # number of capacity decision variables
    prodsize = lpcosts.production.size  # number of production decision variables
    bsize = adapter.b_ub.size  # number of constraints

    assert adapter.c.size == capsize + prodsize
    assert adapter.A_ub.shape[0] == bsize
    assert adapter.A_ub.shape[1] == capsize + prodsize

    assert bsize == lpcosts.replacement.size
    assert adapter.A_ub[:, capsize:] == approx(0)
    assert set(adapter.A_ub[:, :capsize].flatten()) == {0.0, 1.0}


def test_scipy_adapter_no_constraint(lp_inputs, lpcosts):
    adapter = ScipyAdapter.from_muse_data(
        **lp_inputs,
        constraints=[],
    )

    assert set(adapter.kwargs) == {"c", "A_ub", "b_ub", "A_eq", "b_eq", "bounds"}
    assert adapter.bounds == (0, np.inf)
    assert all(
        getattr(adapter, attr) is None for attr in ["A_ub", "b_ub", "A_eq", "b_eq"]
    )
    assert adapter.c.ndim == 1
    assert adapter.c.size == lpcosts.capacity.size + lpcosts.production.size


def test_back_to_muse_quantities(lpcosts):
    data = unified_dataset(lpcosts)

    # Test capacity
    lpquantity = selected_quantity(data, "capacity")
    assert set(lpquantity.dims) == {"d(asset)", "d(replacement)"}
    copy = back_to_muse_quantity(
        lpquantity.costs.values, xr.zeros_like(lpquantity.costs)
    )
    assert (copy == lpcosts.capacity).all()

    # Test production
    lpquantity = selected_quantity(data, "production")
    assert set(lpquantity.dims) == {
        "d(asset)",
        "d(replacement)",
        "d(timeslice)",
        "d(commodity)",
    }
    copy = back_to_muse_quantity(
        lpquantity.costs.values, xr.zeros_like(lpquantity.costs)
    )
    assert (copy == lpcosts.production).all()


def test_back_to_muse_all(lpcosts, lp_inputs):
    data = unified_dataset(lpcosts)
    lpcapacity = selected_quantity(data, "capacity")
    lpproduction = selected_quantity(data, "production")

    x = np.concatenate(
        (
            lpcosts.capacity.transpose(
                *[u[2:-1] for u in lpcapacity.dims]
            ).values.flatten(),
            lpcosts.production.transpose(
                *[u[2:-1] for u in lpproduction.dims]
            ).values.flatten(),
        )
    )

    adapter = ScipyAdapter.from_muse_data(**lp_inputs, constraints=[])
    copy = adapter.to_muse(x)
    assert copy.capacity.size + copy.production.size == x.size
    assert (copy.capacity == lpcosts.capacity).all()
    assert (copy.production == lpcosts.production).all()


def test_scipy_adapter_standard_constraints(lp_inputs, constraints):
    adapter = ScipyAdapter.from_muse_data(
        **lp_inputs, constraints=list(constraints.values())
    )

    n_constraints = adapter.b_ub.size
    n_decision_vars = adapter.c.size
    maxprod_constraint = constraints["max_production"]

    assert (
        n_decision_vars
        == lp_inputs["capacity_costs"].size + maxprod_constraint.production.size
    )
    assert adapter.b_eq is None
    assert adapter.A_eq is None
    assert adapter.A_ub.shape == (n_constraints, n_decision_vars)
    assert n_constraints == sum(c.b.size for c in constraints.values())


def test_scipy_solver(model_data, lp_inputs, constraints):
    """Test the scipy solver for demand matching."""
    from muse.investments import scipy_match_demand

    solution = scipy_match_demand(
        costs=lp_inputs["capacity_costs"],
        commodities=lp_inputs["commodities"],
        search_space=model_data["search_space"],
        technologies=model_data["technologies"],
        constraints=list(constraints.values()),
    )
    assert isinstance(solution, xr.DataArray)
    assert set(solution.dims) == {"asset", "replacement"}



"""Output cached quantities.

Functions that output the state of diverse quantities at intermediate steps of the
calculation.

The core of the method is the OutputCache class that initiated by the MCA with input
parameters defined in the TOML file, much like the existing `output` options but in a
`outputs_cache` list, enables listening for data to be cached and, after
each period, saved into disk via the `consolidate_cache` method.

Anywhere in the code, you can write:

.. code-block:: python

    cache_quantity(quantity_name=some_data)

If the quantity has been set as something to cache, the data will be stored and,
eventually, save to disk after - possibly - aggregating the data and removing those
entries corresponding to non-convergent investment attempts. This process of cleaning
and aggregation is quantity specific.

See documentation for the :py:func:`muse.outputs.cache.cache_quantity` function as well
as how to setup the toml input file to cache quantities. Users can customize and create
further output quantities by registering with MUSE via
:py:func:`muse.outputs.cache.register_cached_quantity`.
"""

from __future__ import annotations

from collections import ChainMap
from collections.abc import Mapping, MutableMapping, Sequence
from functools import reduce
from operator import attrgetter
from typing import (
    Callable,
    Union,
)

import pandas as pd
import xarray as xr
from pubsub import pub

from muse.registration import registrator
from muse.sectors import AbstractSector

OUTPUT_QUANTITY_SIGNATURE = Callable[
    [list[xr.DataArray]], Union[xr.DataArray, pd.DataFrame]
]
"""Signature of functions computing quantities for later analysis."""

OUTPUT_QUANTITIES: MutableMapping[str, OUTPUT_QUANTITY_SIGNATURE] = {}
"""Quantity for post-simulation analysis."""

CACHE_TOPIC_CHANNEL = "cache_quantity"
"""Topic channel to use with the pubsub messaging system."""


@registrator(registry=OUTPUT_QUANTITIES)
def register_cached_quantity(function: OUTPUT_QUANTITY_SIGNATURE) -> Callable:
    """Registers a function to compute an output quantity."""
    from functools import wraps

    @wraps(function)
    def decorated(*args, **kwargs):
        result = function(*args, **kwargs)
        if isinstance(result, (pd.DataFrame, xr.DataArray)):
            result.name = function.__name__
        return result

    return decorated


def cache_quantity(
    function: Callable | None = None,
    quantity: str | Sequence[str] | None = None,
    **kwargs: xr.DataArray,
) -> Callable | None:
    """Cache one or more quantities to be post-processed later on.

    This function can be used as a decorator, in which case the quantity input argument
    must be set, or directly called with any number of keyword arguments. In the former
    case, the matching between quantities and values to cached is done by the function
    'match_quantities'. When used in combination with other decorators, care must be
    taken to decide the order in which they are applied to make sure the appropriate
    output is cached.

    Note that if the quantity has NOT been selected to be cached when configuring the
    MUSE simulation, it will be silently ignored if present as an input to this
    function.

    Example:
        As a decorator, the quantity argument must be set:

        >>> @cache_quantity(quantity="capacity")
        ... def some_calculation():
        ...     return xr.DataArray()

        If returning a sequence of DataArrays, the number of quantities to record must
        be the same as the number of arrays. They are paired in the same order they are
        given and the 'name' attribute of the arrays, if present, is ignored.

        >>> @cache_quantity(quantity=["capacity", "production"])
        ... def other_calculation():
        ...     return xr.DataArray(), xr.DataArray()

        For a finer control of what is cached when there is a complex output, combine
        the DataArrays in a Dataset. In this case, the 'quantity' input argument can be
        either a string or a sequence of strings to record multiple variables in the
        Dataset.

        >>> @cache_quantity(quantity=["capacity", "production"])
        ... def and_another_one():
        ...     return xr.Dataset(
        ...         {
        ...             "not cached": xr.DataArray(),
        ...             "capacity": xr.DataArray(),
        ...             "production": xr.DataArray(),
        ...         }
        ...     )

        When this function is called directly and not used as a decorator, simply
        provide the name of the quantities and the DataArray to record as keyword
        arguments:

        >>> cache_quantity(capacity=xr.DataArray(), production=xr.DataArray())

    Args:
        function (Optional[Callable]): The decorated function, if any. Its output must
            be a DataArray, a sequence of DataArray or a Dataset. See 'match_quantities'
        quantity (Union[str, List[str], None]): The name of the quantities to record.
        **kwargs (xr.DataArray): Keyword arguments of the form
            'quantity_name=quantity_value'.

    Raises:
        ValueError: If a function input argument is provided at the same time than
        keyword arguments.

    Return:
        (Optional[Callable]) The decorated function (or a dummy function if called
        directly).
    """
    from functools import wraps

    # When not used as a decorator
    if len(kwargs) > 0:
        if function is not None:
            raise ValueError(
                "If keyword arguments are provided, then 'function' must be None"
            )
        pub.sendMessage(CACHE_TOPIC_CHANNEL, data=kwargs)
        return None

    # When used as a decorator
    if function is None:
        return lambda x: cache_quantity(x, quantity=quantity)

    if quantity is None:
        raise ValueError(
            "When 'cache_quantity' is used as a decorator the 'quantity' input argument"
            " must be a string or sequence of strings. None found."
        )

    @wraps(function)
    def decorated(*args, **kwargs):
        result = function(*args, **kwargs)
        cache_quantity(**match_quantities(quantity, result))
        return result

    return decorated


def match_quantities(
    quantity: str | Sequence[str],
    data: xr.DataArray | xr.Dataset | Sequence[xr.DataArray],
) -> Mapping[str, xr.DataArray]:
    """Matches the quantities with the corresponding data.

    The possible name attribute in the DataArrays is ignored.

    Args:
        quantity (Union[str, Sequence[str]]): The name(s) of the quantity(ies) to cache.
        data (Union[xr.DataArray, xr.Dataset, Sequence[xr.DataArray]]): The structure
            containing the data to cache.

    Raises:
        TypeError: If there is an invalid combination of input argument types.
        ValueError: If the number of quantities does not match the length of the data.
        KeyError: If the required quantities do not exist as variables in the dataset.

    Returns:
        (Mapping[str, xr.DataArray]) A dictionary matching the quantity names with the
        corresponding data.
    """
    if isinstance(quantity, str) and isinstance(data, xr.DataArray):
        return {quantity: data}

    elif isinstance(quantity, str) and isinstance(data, xr.Dataset):
        return {quantity: data[quantity]}

    elif isinstance(quantity, Sequence) and isinstance(data, xr.Dataset):
        return {q: data[q] for q in quantity}

    elif isinstance(quantity, Sequence) and isinstance(data, Sequence):
        if len(quantity) != len(data):
            msg = f"{len(quantity)} != {len(data)}"
            raise ValueError(
                f"The number of quantities does not match the length of the data {msg}."
            )
        return {q: v for q, v in zip(quantity, data)}

    else:
        msg = f"{type(quantity)} and {type(data)}"
        raise TypeError(f"Invalid combination of input argument types {msg}")


class OutputCache:
    """Creates outputs functions for post-mortem analysis of cached quantities.

    Each parameter is a dictionary containing the following:

    - quantity (mandatory): name of the quantity to output. Mandatory.
    - sink (optional): name of the storage procedure, e.g. the file format
      or database format. When it cannot be guessed from `filename`, it defaults to
      "csv".
    - filename (optional): path to a directory or a file where to store the quantity. In
      the latter case, if sink is not given, it will be determined from the file
      extension. The filename can incorporate markers. By default, it is
      "{default_output_dir}/{sector}{year}{quantity}{suffix}".
    - any other parameter relevant to the sink, e.g. `pandas.to_csv` keyword
      arguments.

    For simplicity, it is also possible to given lone strings as input.
    They default to `{'quantity': string}` (and the sink will default to
    "csv").

    Raises:
        ValueError: If unknown quantities are requested to be cached.
    """

    def __init__(
        self,
        *parameters: Mapping,
        output_quantities: MutableMapping[str, OUTPUT_QUANTITY_SIGNATURE] | None = None,
        sectors: list[AbstractSector] | None = None,
        topic: str = CACHE_TOPIC_CHANNEL,
    ):
        from muse.outputs.sector import _factory

        output_quantities = (
            OUTPUT_QUANTITIES if output_quantities is None else output_quantities
        )
        self.agents: MutableMapping[str, MutableMapping[str, str]] = (
            extract_agents(sectors) if sectors is not None else {}
        )

        missing = [
            p["quantity"] for p in parameters if p["quantity"] not in output_quantities
        ]

        if len(missing) != 0:
            raise ValueError(
                f"There are unknown quantities to cache: {missing}. "
                f"Valid quantities are: {list(output_quantities.keys())}"
            )

        self.to_save: Mapping[str, list[xr.DataArray]] = {
            p["quantity"]: [] for p in parameters if p["quantity"] in output_quantities
        }

        self.factory: Mapping[str, Callable] = {
            p["quantity"]: _factory(output_quantities, p, sector_name="Cache")
            for p in parameters
            if p["quantity"] in self.to_save
        }

        self.normalized = {}
        for p in self.to_save:
            for alt, val in output_quantities.items():
                if val != output_quantities[p]:
                    continue
                self.normalized[alt] = p

        pub.subscribe(self.cache, topic)

    def cache(self, data: Mapping[str, xr.DataArray]) -> None:
        """Caches the data into memory.

        If the quantity has not been selected to be cached when configuring the
        MUSE simulation, it will be silently ignored if present as an input to this
        function.

        Args:
            data (Mapping[str, xr.DataArray]): Dictionary with the quantities and
            DataArray values to save.
        """
        for quantity, value in data.items():
            normalized = self.normalized.get(quantity, "")
            if normalized not in self.to_save:
                continue

            self.to_save[normalized].append(value.copy())
            self.to_save[normalized][-1].name = normalized

    def consolidate_cache(self, year: int) -> None:
        """Save the cached data into disk and flushes cache.

        This method is meant to be called after each time period in the main loop of the
        MCA, just after market and sector quantities are saved.

        Args:
            year (int): Year of interest.
        """
        for quantity, cache in self.to_save.items():
            if len(cache) == 0:
                continue

            self.factory[quantity](cache, self.agents, year=year)
        self.to_save = {q: [] for q in self.to_save}


def extract_agents(
    sectors: list[AbstractSector],
) -> MutableMapping[str, MutableMapping[str, str]]:
    """_summary_.

    Args:
        sectors (List[AbstractSector]): _description_

    Returns:
        Mapping[Text, Text]: _description_
    """
    return ChainMap(*[extract_agents_internal(sector) for sector in sectors])


def extract_agents_internal(
    sector: AbstractSector,
) -> MutableMapping[str, MutableMapping[str, str]]:
    """Extract simple agent metadata from a sector.

    Args:
        sector (AbstractSector): Sector to extract the metadata from.

    Returns:
        Mapping[Text, Text]: A dictionary with the uuid of each agent as keys and a
        dictionary with the name, agent type and agent sector as values.
    """
    info: MutableMapping[str, MutableMapping[str, str]] = {}
    sector_name = getattr(sector, "name", "unnamed")
    agents = sorted(getattr(sector, "agents", []), key=attrgetter("name"))
    for agent in agents:
        aid = agent.uuid
        info[aid] = {}
        info[aid]["agent"] = agent.name
        info[aid]["category"] = agent.category
        info[aid]["sector"] = sector_name
        info[aid]["dst_region"] = agent.region
    return info


def _aggregate_cache(quantity: str, data: list[xr.DataArray]) -> pd.DataFrame:
    """Combine a list of DataArrays in a dataframe.

    The merging gives precedence to the last entries of the list over the first ones.
    I.e, the records of the arrays cached last will overwrite those of the ones cached
    before in the case of having identical index.

    Args:
        quantity: The quantity to cache.
        data: The list of DataArrays to combine.

    Returns:
        pd.DataFrame: A Dataframe with the data aggregated.
    """
    data = [da.to_dataframe().reset_index() for da in data]
    cols = sorted(
        set.intersection(
            *map(set, [[c for c in d.columns if c != quantity] for d in data])
        )
    )

    def check_col(colname: str) -> str:
        if colname.endswith("_x") or colname.endswith("_y"):
            return colname.rsplit("_", maxsplit=1)[0]
        return colname

    return reduce(
        lambda left, right: pd.DataFrame.merge(left, right, how="outer", on=cols)
        .T.groupby(check_col)
        .last()
        .T,
        data,
    )


def consolidate_quantity(
    quantity: str,
    cached: list[xr.DataArray],
    agents: MutableMapping[str, MutableMapping[str, str]],
) -> pd.DataFrame:
    """Consolidates the cached quantity into a single DataFrame to save.

    Args:
        quantity (Text): The quantity to cache.
        cached (List[xr.DataArray]): The list of cached arrays
        agents (MutableMapping[Text, MutableMapping[Text, Text]]): Agents' metadata.

    Returns:
        pd.DataFrame: DataFrame with the consolidated data.
    """
    data = _aggregate_cache(quantity, cached)

    ignore_dst_region = "dst_region" in data.columns
    for agent in tuple(agents):
        filter = data.agent == agent
        for key, value in agents[agent].items():
            if key == "dst_region" and ignore_dst_region:
                continue
            data.loc[filter, key] = value

    return data[sorted(data.columns)]


@register_cached_quantity
def capacity(
    cached: list[xr.DataArray],
    agents: MutableMapping[str, MutableMapping[str, str]],
    **kwargs,
) -> pd.DataFrame:
    """Consolidates the cached capacities into a single DataFrame to save.

    Args:
        cached (List[xr.DataArray]): The list of cached arrays
        agents (MutableMapping[Text, MutableMapping[Text, Text]]): Agents' metadata.
        kwargs: unused.

    Returns:
        pd.DataFrame: DataFrame with the consolidated data.
    """
    return consolidate_quantity("capacity", cached, agents)


@register_cached_quantity
def production(
    cached: list[xr.DataArray],
    agents: MutableMapping[str, MutableMapping[str, str]],
    **kwargs,
) -> pd.DataFrame:
    """Consolidates the cached production into a single DataFrame to save.

    Args:
        cached (List[xr.DataArray]): The list of cached arrays
        agents (MutableMapping[Text, MutableMapping[Text, Text]]): Agents' metadata.
        kwargs: unused.

    Returns:
        pd.DataFrame: DataFrame with the consolidated data.
    """
    return consolidate_quantity("production", cached, agents)


@register_cached_quantity(name="lifetime_levelized_cost_of_energy")
def lcoe(
    cached: list[xr.DataArray],
    agents: MutableMapping[str, MutableMapping[str, str]],
    **kwargs,
) -> pd.DataFrame:
    """Consolidates the cached LCOE into a single DataFrame to save.

    Args:
        cached (List[xr.DataArray]): The list of cached arrays
        agents (MutableMapping[Text, MutableMapping[Text, Text]]): Agents' metadata.
        kwargs: unused.

    Returns:
        pd.DataFrame: DataFrame with the consolidated data.
    """
    """Consolidates the cached LCOE into a single DataFrame to save."""
    if "timeslice" in cached[0].dims:
        cached = [c.assign_coords(timeslice=c.timeslice.data) for c in cached]
    return consolidate_quantity("lcoe", cached, agents)



import numpy as np
import xarray as xr
from pytest import mark

from muse.investments import cliff_retirement_profile


def test_cliff_retirement_known_profile():
    """Test cliff retirement profile with known values and expected output."""
    technology = ["a", "b", "c"]
    lifetime = xr.DataArray(
        np.arange(1, len(technology) + 1),
        dims="technology",
        coords={"technology": technology},
        name="technical_life",
    )

    profile = cliff_retirement_profile(technical_life=lifetime, investment_year=2020)
    expected = np.array(
        [
            [True, False, False, False],
            [True, True, False, False],
            [True, True, True, False],
        ]
    )

    assert set(profile.dims) == {"year", "technology"}
    assert (profile == expected.T).all()


@mark.parametrize("protected", range(12))
def test_cliff_retirement_random_profile(protected):
    """Test cliff retirement profile with random lifetimes and protected periods."""
    technology = list("abcde")
    lifetime = xr.DataArray(
        sorted(np.random.randint(1, 10, len(technology))),
        dims="technology",
        coords={"technology": technology},
        name="technical_life",
    )

    investment_year = 2020
    effective_lifetime = (protected // lifetime + 1) * lifetime
    profile = cliff_retirement_profile(
        technical_life=lifetime.clip(min=protected), investment_year=investment_year
    )

    # Verify profile boundaries and properties
    profile_int = profile.astype(int)
    assert profile.year.min() == investment_year
    assert profile.year.max() <= investment_year + effective_lifetime.max() + 1
    assert profile_int.interp(year=investment_year).all()
    assert profile_int.interp(year=investment_year + protected - 1).all()
    assert not profile_int.interp(year=profile.year.max()).any()



"""Modes of interactions between agents.

Interactions between agents are modelled via two orthogonal concepts:

- a *net* is a set of agents which interact in some way
- an *interaction* proper is a function that takes a net and actually performs the
  interaction.

Hence, there are two registrators in this this module,
:py:func:`register_interaction_net`, and :py:func:`register_agent_interaction`. The
first registers functions that take the full set of agents as input and returns a
sequence of nets. It is expected each net of the sequence will be applied the same
interaction. The second registrator registers the interaction proper: it takes agents as
arguments and returns nothing. It is expected to modify the agents in-place.
"""

from __future__ import annotations

__all__ = [
    "factory",
    "new_to_retro_net",
    "register_agent_interaction",
    "register_interaction_net",
    "transfer_assets",
]

from collections.abc import Mapping, Sequence
from typing import Callable

from muse.agents import AbstractAgent, Agent
from muse.errors import NoInteractionsFound
from muse.registration import registrator

AGENT_INTERACTIONS: Mapping[str, Callable] = {}
"""All interaction between a single agents and its interactees."""
INTERACTION_NETS: Mapping[str, Callable] = {}
"""All functions to computes lists of agents interaction with each other."""

INTERACTION_NET = Sequence[Sequence[Agent]]
"""Defines interaction sets between agents.

An interaction set is a mapping an agent to the agents with which it interacts. For
instance, it could, for each region, map a new agent with the retro agent with the same
name as itself.
"""
INTERACTION_NET_SIGNATURE = Callable[[Sequence[Agent]], INTERACTION_NET]
"""Signature of the interaction net function.

The interaction net functions should create a list of lists of interacting agents, given
as input the list of all agents.
"""
AGENT_INTERACTION_SIGNATURE = Callable[[Agent, Agent], None]
"""Signature of a single agent to agent(s) interaction.

An interaction function takes as argument the agents that are to interact. It can modify
the parameters and assets of these agents.
"""


@registrator(registry=INTERACTION_NETS, loglevel="debug")
def register_interaction_net(function: INTERACTION_NET_SIGNATURE):
    """Decorator to register a function computing interaction nets.

    An interaction net function takes as input the list of all agents and
    returns the list of all interactions, where an interaction is a list of at
    least two interacting agents.

    An interactiont-net function also takes as argument a sector object.
    This object should not be modified in any way. But it can be queried for
    parameters, if the specific interaction-net function requires it.
    """
    return function


@registrator(registry=AGENT_INTERACTIONS, loglevel="debug")
def register_agent_interaction(function: AGENT_INTERACTION_SIGNATURE):
    """Decorator to register an agent to agent(s) interaction function.

    An agent interaction function takes at least two agents and makes them
    interact in some way.

    An agent interaction function also takes as argument a sector object.
    This object should not be modified in any way. But it can be queried for
    parameters, if the specific agent interaction function requires it. This is
    most likely the same configuration object passed on to the interaction net
    function.
    """
    return function


def factory(
    inputs: Sequence[Mapping | tuple[str, str]] | None = None,
) -> Callable[[Sequence[AbstractAgent]], None]:
    """Creates an interaction functor."""
    if inputs is None:
        inputs = tuple()

    interactions = []
    for params in inputs:
        if isinstance(params, Mapping):
            net_params = params["net"]
            action_params = params["interaction"]
        else:
            net_params, action_params = params

        if isinstance(net_params, str):
            net_params = {"name": net_params}

        if isinstance(action_params, str):
            action_params = {"name": action_params}

        net = INTERACTION_NETS[net_params["name"]]
        interaction = AGENT_INTERACTIONS[action_params["name"]]

        interactions.append(((net, net_params), (interaction, action_params)))

    def compute_interactions(agents: Sequence[AbstractAgent]) -> None:
        """Applies interaction net and agent interaction functions.

        If a network is found with no interactions, an error is raised.
        """
        from logging import getLogger

        for (net, net_params), (interaction, interaction_params) in interactions:
            nparams = {k: v for k, v in net_params.items() if k != "name"}
            sets = net(agents, **nparams)

            if len(sets) == 0:
                raise NoInteractionsFound

            getLogger(__name__).info(
                f"Net {net_params['name']} of {len(sets)} interactions interacting "
                f"via {interaction_params['name']}"
            )

            iparams = {k: v for k, v in interaction_params.items() if k != "name"}
            for agents in sets:
                interaction(*agents, **iparams)

    return compute_interactions


def agents_groupby(
    agents: Sequence[Agent], attributes: Sequence[str]
) -> Mapping[tuple, list[Agent]]:
    attr_list = [tuple(getattr(agent, attr) for attr in attributes) for agent in agents]
    result: Mapping[tuple, list[Agent]] = {tuple(n): [] for n in attr_list}
    for attrs, agent in zip(attr_list, agents):
        result[attrs].append(agent)
    return result


@register_interaction_net(name=["default", "new_to_retro"])
def new_to_retro_net(
    agents: Sequence[Agent], first_category: str = "newcapa"
) -> INTERACTION_NET:
    """Interactions between new and retrofit agents."""
    groups = agents_groupby(agents, ("region", "name"))

    def comparison(a, b):
        if a is None:
            return b is None
        if a == b:
            return True
        if hasattr(a, "lower") and hasattr(b, "lower"):
            return a.lower() == b.lower()
        return False

    return [
        [a for a in group if comparison(a.category, first_category)]
        + [a for a in group if not comparison(a.category, first_category)]
        for group in groups.values()
        if len(group) == 2
    ]


@register_agent_interaction(name="transfer")
def transfer_assets(from_: Agent, to_: Agent) -> None:
    """Transfer assets from first agent to second agent."""
    from xarray import zeros_like

    from muse.utilities import merge_assets

    to_.assets = merge_assets(to_.assets, from_.assets)
    from_.assets = zeros_like(to_.assets)



import os
import shutil
from pathlib import Path

import pandas as pd

from muse import examples

parent_path = Path(__file__).parent


def generate_model_1() -> None:
    """Generates the first model for tutorial 7.

    Adds a minimum service factor constraint for gasCCGT.

    """
    model_name = "1-min-constraint"
    model_path = parent_path / model_name
    if model_path.exists():
        shutil.rmtree(model_path)

    # Starting point: copy default_timeslice model
    examples.copy_model(name="default_timeslice", path=parent_path, overwrite=True)
    os.rename(parent_path / "model", model_path)

    # Modify MinimumServiceFactor for gasCCGT
    timeslices_file = model_path / "power/TechnodataTimeslices.csv"
    df = pd.read_csv(timeslices_file)
    df.loc[:, "minimum_service_factor"] = 0
    df.loc[df["technology"] == "gasCCGT", "minimum_service_factor"] = [
        0.2,
        0.4,
        0.6,
        0.4,
        0.8,
        1,
    ]
    df.to_csv(timeslices_file, index=False)


def generate_model_2() -> None:
    """Generates the second model for tutorial 7.

    Adds a maximum service factor constraint for windturbine.

    """
    model_name = "2-max-constraint"
    model_path = parent_path / model_name
    if model_path.exists():
        shutil.rmtree(model_path)

    # Starting point: copy model from 1-min-constraint
    shutil.copytree(parent_path / "1-min-constraint", model_path)
    if (model_path / "Results").exists():
        shutil.rmtree(model_path / "Results")

    # Modify UtilizationFactor and MinimumServiceFactor for windturbine
    timeslices_file = model_path / "power/TechnodataTimeslices.csv"
    df = pd.read_csv(timeslices_file)
    df.loc[df["technology"] == "gasCCGT", "minimum_service_factor"] = [
        0.2,
        0.4,
        0.6,
        0.4,
        0,
        0,
    ]
    df.loc[df["technology"] == "gasCCGT", "utilization_factor"] = [
        1,
        1,
        1,
        1,
        0.5,
        0.5,
    ]
    df.to_csv(timeslices_file, index=False)


if __name__ == "__main__":
    generate_model_1()
    generate_model_2()



"""Test timeslice utilities."""

import numpy as np
from pytest import approx, fixture, raises
from xarray import DataArray

from muse.timeslices import (
    broadcast_timeslice,
    compress_timeslice,
    distribute_timeslice,
    drop_timeslice,
    expand_timeslice,
    get_level,
    read_timeslices,
    sort_timeslices,
    timeslice_max,
)

# Constants
TIMESLICE_LEVELS = ["month", "day", "hour"]
TIMESLICE_COORDS = {"timeslice", "month", "day", "hour"}


@fixture
def non_timesliced_dataarray():
    """Create a simple non-timesliced DataArray for testing."""
    return DataArray([1, 2, 3], dims=["x"])


@fixture
def timeslice_toml():
    """TOML configuration for timeslice testing."""
    return """
    [timeslices]
    winter.weekday.night = 396
    winter.weekday.morning = 396
    winter.weekday.afternoon = 264
    winter.weekend.night = 156
    winter.weekend.morning = 156
    winter.weekend.afternoon = 156
    springautumn.weekday.night = 792
    springautumn.weekday.morning = 792
    springautumn.weekday.afternoon = 528
    springautumn.weekend.night = 300
    springautumn.weekend.morning = 300
    springautumn.weekend.afternoon = 300
    summer.weekday.night = 396
    summer.weekday.morning  = 396
    summer.weekday.afternoon = 264
    summer.weekend.night = 150
    summer.weekend.morning = 150
    summer.weekend.afternoon = 150
    """


@fixture(params=[False, True])
def timeslice(request, timeslice_toml):
    """Generate a timeslice DataArray.

    Creates two versions:
    - A one-dimensional DataArray with a single "timeslice" dimension
    - A two-dimensional DataArray with an additional "x" dimension and randomized
     weights
    """
    from toml import loads

    ts = read_timeslices(loads(timeslice_toml))

    if request.param:
        ts = ts.expand_dims({"x": 3})
        ts = ts + np.random.randint(0, 10, ts.shape)

    return ts


def test_no_overlap():
    """Test that overlapping timeslice definitions raise ValueError."""
    invalid_toml = """
    [timeslices]
    winter.weekday.night = 396
    winter.weekday.morning = 396
    winter.weekday.weekend = 156
    winter.weekend.night = 156
    winter.weekend.morning = 156
    winter.weekend.weekend = 156
    """
    with raises(ValueError):
        read_timeslices(invalid_toml)


def test_drop_timeslice(non_timesliced_dataarray, timeslice):
    """Test dropping timeslice coordinates."""
    timesliced = broadcast_timeslice(non_timesliced_dataarray, ts=timeslice)
    dropped = drop_timeslice(timesliced)

    assert TIMESLICE_COORDS.issubset(timesliced.coords)
    assert not TIMESLICE_COORDS.intersection(dropped.coords)
    assert drop_timeslice(non_timesliced_dataarray).equals(non_timesliced_dataarray)
    assert drop_timeslice(dropped).equals(dropped)


def test_broadcast_timeslice(non_timesliced_dataarray, timeslice):
    """Test broadcasting arrays to different timeslice granularities."""
    for level in TIMESLICE_LEVELS:
        out = broadcast_timeslice(non_timesliced_dataarray, ts=timeslice, level=level)
        target = compress_timeslice(
            timeslice, ts=timeslice, level=level, operation="sum"
        )

        assert out.timeslice.equals(target.timeslice)
        assert (out.diff(dim="timeslice") == 0).all()
        assert all(
            (out.isel(timeslice=i) == non_timesliced_dataarray).all()
            for i in range(out.sizes["timeslice"])
        )

    # Test broadcasting an already timesliced array (should return unchanged)
    out2 = broadcast_timeslice(out, ts=timeslice)
    assert out2.equals(out)

    # Test broadcasting with incompatible timeslice levels
    with raises(ValueError):
        broadcast_timeslice(
            compress_timeslice(out, ts=timeslice, level="day"), ts=timeslice
        )


def test_distribute_timeslice(non_timesliced_dataarray, timeslice):
    """Test distributing arrays across timeslices."""
    for level in TIMESLICE_LEVELS:
        out = distribute_timeslice(non_timesliced_dataarray, ts=timeslice, level=level)
        target = compress_timeslice(
            timeslice, ts=timeslice, level=level, operation="sum"
        )

        assert out.timeslice.equals(target.timeslice)

        # Check proportionality
        out_prop = out / broadcast_timeslice(
            out.sum("timeslice"), ts=timeslice, level=level
        )
        ts_prop = target / broadcast_timeslice(
            target.sum("timeslice"), ts=timeslice, level=level
        )
        assert abs(out_prop - ts_prop).max() < 1e-6

        assert (out.sum("timeslice") == approx(non_timesliced_dataarray)).all()

    # Test distributing an already timesliced array (should return unchanged)
    assert distribute_timeslice(out, ts=timeslice).equals(out)

    # Test distributing with incompatible timeslice levels
    with raises(ValueError):
        distribute_timeslice(
            compress_timeslice(out, ts=timeslice, level="day"), ts=timeslice
        )


def test_compress_timeslice(non_timesliced_dataarray, timeslice):
    """Test compressing timesliced arrays."""
    timesliced = broadcast_timeslice(non_timesliced_dataarray, ts=timeslice)

    for level in TIMESLICE_LEVELS:
        for operation in ["sum", "mean"]:
            out = compress_timeslice(
                timesliced, ts=timeslice, operation=operation, level=level
            )
            assert get_level(out) == level

            if operation == "sum":
                assert (
                    out.sum("timeslice") == approx(timesliced.sum("timeslice"))
                ).all()
            else:  # mean
                assert (
                    out.mean("timeslice") == approx(timesliced.mean("timeslice"))
                ).all()

    # Test compressing without specifying a level (should return unchanged)
    assert compress_timeslice(timesliced, ts=timeslice).equals(timesliced)

    # Test compressing with invalid level name
    with raises(ValueError):
        compress_timeslice(timesliced, ts=timeslice, level="invalid")

    # Test compressing with invalid operation type
    with raises(ValueError):
        compress_timeslice(timesliced, ts=timeslice, level="day", operation="invalid")


def test_expand_timeslice(non_timesliced_dataarray, timeslice):
    """Test expanding timesliced arrays."""
    for level in TIMESLICE_LEVELS:
        timesliced = broadcast_timeslice(
            non_timesliced_dataarray, ts=timeslice, level=level
        )

        for operation in ["broadcast", "distribute"]:
            out = expand_timeslice(timesliced, ts=timeslice, operation=operation)
            assert out.timeslice.equals(timeslice.timeslice)

            if operation == "broadcast":
                assert (
                    out.mean("timeslice") == approx(timesliced.mean("timeslice"))
                ).all()
            else:  # distribute
                assert (
                    out.sum("timeslice") == approx(timesliced.sum("timeslice"))
                ).all()

    # Test expanding an already expanded array (should return unchanged)
    assert expand_timeslice(out, ts=timeslice).equals(out)

    # Test expanding with invalid operation type
    timesliced = broadcast_timeslice(
        non_timesliced_dataarray, ts=timeslice, level="month"
    )
    with raises(ValueError):
        expand_timeslice(timesliced, ts=timeslice, operation="invalid")


def test_get_level(non_timesliced_dataarray, timeslice):
    """Test getting timeslice level."""
    for level in TIMESLICE_LEVELS:
        timesliced = broadcast_timeslice(
            non_timesliced_dataarray, ts=timeslice, level=level
        )
        assert get_level(timesliced) == level

    with raises(ValueError):
        get_level(non_timesliced_dataarray)


def test_sort_timeslices(non_timesliced_dataarray, timeslice):
    """Test sorting timeslices."""
    # Test hour level
    timesliced = broadcast_timeslice(
        non_timesliced_dataarray, ts=timeslice, level="hour"
    )
    sorted_data = sort_timeslices(timesliced, timeslice)
    assert sorted_data.timeslice.equals(timeslice.timeslice)

    # Test month level
    timesliced = broadcast_timeslice(
        non_timesliced_dataarray, ts=timeslice, level="month"
    )
    sorted_data = sort_timeslices(timesliced, timeslice)
    assert sorted_data.timeslice.equals(timesliced.sortby("timeslice").timeslice)


def test_timeslice_max(non_timesliced_dataarray):
    """Test timeslice maximum calculation."""
    ts = read_timeslices("""
        [timeslices]
        winter.weekday.night = 396
        winter.weekday.morning = 396
    """)

    timesliced = broadcast_timeslice(non_timesliced_dataarray, ts=ts)
    timesliced = timesliced + np.random.rand(*timesliced.shape)
    max_val = timeslice_max(timesliced, ts=ts)
    assert max_val.equals(timesliced.max("timeslice") * 2)



from pytest import fixture, mark

YEAR = 2030


@fixture
def _demand(demand_share):
    return demand_share


@fixture
def _technologies(technologies, demand_share):
    from muse.utilities import broadcast_over_assets

    techs = technologies.sel(year=YEAR).rename(technology="replacement")
    return broadcast_over_assets(techs, demand_share)


@fixture
def _prices(market, demand_share):
    from muse.utilities import broadcast_over_assets

    prices = market.prices.sel(year=YEAR)
    return broadcast_over_assets(prices, demand_share, installed_as_year=False)


def test_fixtures(_technologies, _demand, _prices):
    """Validating that the fixtures have appropriate dimensions."""
    assert set(_technologies.dims) == {"asset", "commodity", "replacement"}
    assert set(_demand.dims) == {"asset", "commodity", "timeslice"}
    assert set(_prices.dims) == {"asset", "commodity", "timeslice"}


@mark.usefixtures("save_registries")
def test_objective_registration():
    from muse.objectives import OBJECTIVES, register_objective

    @register_objective
    def a_objective(*args, **kwargs):
        pass

    assert "a_objective" in OBJECTIVES
    assert OBJECTIVES["a_objective"] is a_objective

    @register_objective(name="something")
    def b_objective(*args, **kwargs):
        pass

    assert "something" in OBJECTIVES
    assert OBJECTIVES["something"] is b_objective


@mark.usefixtures("save_registries")
def test_computing_objectives(_technologies, _demand, _prices):
    from muse.objectives import factory, register_objective

    @register_objective
    def first(technologies, demand, switch=True, *args, **kwargs):
        from xarray import broadcast, full_like

        value = 1 if switch else 2
        result = full_like(
            broadcast(technologies["replacement"], demand["asset"])[0],
            value,
            dtype=float,
        )
        return result

    @register_objective
    def second(technologies, demand, assets=None, *args, **kwargs):
        from xarray import broadcast, full_like

        result = full_like(
            broadcast(technologies["replacement"], demand["asset"])[0], 5, dtype=float
        )
        result[{"asset": assets}] = 3
        return result

    # Test first objective with/without switch
    objectives = factory("first")(
        technologies=_technologies, demand=_demand, prices=_prices, switch=True
    )
    assert set(objectives.data_vars) == {"first"}
    assert (objectives.first == 1).all()
    objectives = factory("first")(
        technologies=_technologies, demand=_demand, prices=_prices, switch=False
    )
    assert (objectives.first == 2).all()

    # Test multiple objectives
    objectives = factory(["first", "second"])(
        technologies=_technologies,
        demand=_demand,
        prices=_prices,
        switch=False,
        assets=0,
    )
    assert set(objectives.data_vars) == {"first", "second"}
    assert (objectives.first == 2).all()
    if len(objectives.asset) > 0:
        assert (objectives.second.isel(asset=0) == 3).all()
    if len(objectives.asset) > 1:
        assert (objectives.second.isel(asset=1) == 5).all()


def test_comfort(_technologies, _demand):
    from muse.objectives import comfort

    _technologies["comfort"] = add_var(_technologies, "replacement")
    result = comfort(_technologies, _demand)
    assert set(result.dims) == {"replacement", "asset"}


def test_efficiency(_technologies, _demand):
    from muse.objectives import efficiency

    _technologies["efficiency"] = add_var(_technologies, "replacement")
    result = efficiency(_technologies, _demand)
    assert set(result.dims) == {"replacement", "asset"}


def test_capacity_to_service_demand(_technologies, _demand):
    from muse.objectives import capacity_to_service_demand

    result = capacity_to_service_demand(_technologies, _demand)
    assert set(result.dims) == {"replacement", "asset"}


def test_capacity_in_use(_technologies, _demand):
    from muse.objectives import capacity_in_use

    result = capacity_in_use(_technologies, _demand)
    assert set(result.dims) == {"replacement", "asset"}


def test_consumption(_technologies, _demand, _prices):
    from muse.objectives import consumption

    result = consumption(_technologies, _demand, _prices)
    assert set(result.dims) == {"replacement", "asset", "timeslice"}


def test_fixed_costs(_technologies, _demand):
    from muse.objectives import fixed_costs

    result = fixed_costs(_technologies, _demand)
    assert set(result.dims) == {"replacement", "asset"}


def test_capital_costs(_technologies, _demand):
    from muse.objectives import capital_costs

    result = capital_costs(_technologies, _demand)
    assert set(result.dims) == {"replacement", "asset"}


def test_emission_cost(_technologies, _demand, _prices):
    from muse.objectives import emission_cost

    result = emission_cost(_technologies, _demand, _prices)
    assert set(result.dims) == {"replacement", "asset", "timeslice"}


def test_fuel_consumption_cost(_technologies, _demand, _prices):
    from muse.objectives import fuel_consumption_cost

    result = fuel_consumption_cost(_technologies, _demand, _prices)
    assert set(result.dims) == {"replacement", "asset", "timeslice"}


def test_annual_levelized_cost_of_energy(_technologies, _demand, _prices):
    from muse.objectives import annual_levelized_cost_of_energy

    result = annual_levelized_cost_of_energy(_technologies, _demand, _prices)
    assert set(result.dims) == {"replacement", "asset"}


def test_lifetime_levelized_cost_of_energy(_technologies, _demand, _prices):
    from muse.objectives import lifetime_levelized_cost_of_energy

    result = lifetime_levelized_cost_of_energy(_technologies, _demand, _prices)
    assert set(result.dims) == {"replacement", "asset"}


def test_net_present_value(_technologies, _demand, _prices):
    from muse.objectives import net_present_value

    result = net_present_value(_technologies, _demand, _prices)
    assert set(result.dims) == {"replacement", "asset"}


def test_net_present_cost(_technologies, _demand, _prices):
    from muse.objectives import net_present_cost

    result = net_present_cost(_technologies, _demand, _prices)
    assert set(result.dims) == {"replacement", "asset"}


def test_equivalent_annual_cost(_technologies, _demand, _prices):
    from muse.objectives import equivalent_annual_cost

    result = equivalent_annual_cost(_technologies, _demand, _prices)
    assert set(result.dims) == {"replacement", "asset"}


def add_var(coordinates, *dims, factor=100.0):
    from numpy.random import rand

    shape = tuple(len(coordinates[u]) for u in dims)
    return dims, (rand(*shape) * factor).astype(type(factor))



"""Aggregates methods to read data from file."""

from muse.defaults import DATA_DIRECTORY
from muse.readers.csv import *  # noqa: F403
from muse.readers.toml import read_settings  # noqa: F401

DEFAULT_SETTINGS_PATH = DATA_DIRECTORY / "default_settings.toml"
"""Default settings path."""



from collections.abc import Mapping
from typing import Any

from pytest import approx, fixture

from muse import constraints as cs
from muse import examples
from muse.utilities import agent_concatenation, reduce_assets


@fixture
def constraints_args(sector="power", model="trade") -> Mapping[str, Any]:
    power = examples.sector(model=model, sector=sector)
    search_space = examples.search_space("power", model="trade")
    market = examples.matching_market("power", "trade")
    assets = agent_concatenation({agent.uuid: agent.assets for agent in power.agents})
    capacity = reduce_assets(assets.capacity, coords=("region", "technology"))

    return dict(
        demand=market.consumption.sel(year=2025, drop=True),
        capacity=capacity.sel(year=[2020, 2025]),
        search_space=search_space,
        market=market,
        technologies=power.technologies.sel(year=2025, drop=True),
    )


def test_demand_constraint(constraints_args):
    constraint = cs.demand(**constraints_args)
    assert set(constraint.b.dims) == {"timeslice", "dst_region", "commodity"}


def test_max_capacity_constraints(constraints_args):
    constraint = cs.max_capacity_expansion(**constraints_args)
    assert constraint.production == 0
    assert set(constraint.capacity.dims) == {"agent", "src_region"}
    assert set(constraint.b.dims) == {"replacement", "dst_region", "src_region"}
    assert set(constraint.agent.coords) == {"region", "agent"}
    assert ((constraint.region == constraint.src_region) == constraint.capacity).all()


def test_max_production(constraints_args):
    constraint = cs.max_production(**constraints_args)
    production_dims = {
        "timeslice",
        "commodity",
        "replacement",
        "agent",
        "dst_region",
        "src_region",
    }
    assert set(constraint.capacity.dims) == production_dims
    assert set(constraint.production.dims) == production_dims
    assert set(constraint.agent.coords) == {"region", "agent"}


def test_minimum_service(constraints_args):
    assert cs.minimum_service(**constraints_args) is None

    constraints_args["technologies"]["minimum_service_factor"] = 0.5
    constraint = cs.minimum_service(**constraints_args)
    service_dims = {"replacement", "agent", "commodity", "timeslice"}
    assert set(constraint.capacity.dims) == service_dims
    assert set(constraint.production.dims) == service_dims
    assert set(constraint.b.dims) == service_dims
    assert set(constraint.agent.coords) == {"region", "agent"}
    assert (constraint.capacity <= 0).all()


def test_search_space(constraints_args):
    search_space = constraints_args["search_space"]
    search_space[:] = 1
    assert cs.search_space(**constraints_args) is None

    search_space[:] = search_space * (search_space.region == "R1")
    constraint = cs.search_space(**constraints_args)
    assert constraint.b.values == approx(0)
    assert constraint.production == 0
    assert set(constraint.b.dims) == {"replacement", "agent"}
    assert set(constraint.capacity.dims) == {"replacement", "agent"}
    assert set(constraint.agent.coords) == {"region", "agent"}


def get_agent_capacities(sector):
    """Helper to get concatenated agent capacities."""
    return agent_concatenation(
        {agent.uuid: agent.assets.capacity for agent in sector.agents}
    )


def test_power_sector_no_investment():
    power = examples.sector("power", "trade")
    market = examples.matching_market("power", "trade").sel(year=[2020, 2025])

    initial_capacity = get_agent_capacities(power)
    power.next(market)
    final_capacity = get_agent_capacities(power)

    assert (initial_capacity == final_capacity).all()


def test_power_sector_some_investment():
    power = examples.sector("power", "trade")
    market = examples.matching_market("power", "trade").sel(year=[2020, 2025])
    market.consumption[:] *= 1.5

    initial_capacity = get_agent_capacities(power)
    result = power.next(market)
    final_capacity = get_agent_capacities(power)

    assert "windturbine" not in initial_capacity.technology
    assert (
        final_capacity.sel(
            asset=final_capacity.technology == "windturbine", year=2025
        ).sum()
        < 1
    )
    assert "dst_region" not in result.dims



import os
from concurrent.futures import ProcessPoolExecutor
from pathlib import Path
from typing import Union

from muse.mca import MCA
from muse.readers.toml import read_settings

parent_path = Path(__file__).parent


def run_model(settings_file: Path) -> Union[str, None]:
    """Run a model with the specified settings file.

    Args:
        settings_file: Path to the settings file to run the model with.

    Returns:
        Error message for failed runs, otherwise None.
    """
    try:
        os.chdir(settings_file.parent)
        settings = read_settings(settings_file)
        model = MCA.factory(settings)
        model.run()
        return None
    except Exception as e:
        return f"Failed to run model with settings file {settings_file}. Error: {e}"


if __name__ == "__main__":
    with ProcessPoolExecutor() as executor:
        futures = {
            executor.submit(run_model, settings_file)
            for settings_file in parent_path.rglob("*/*settings.toml")
        }
    for future in futures:
        error_message = future.result()
        if error_message:
            print(error_message)



import os
import shutil
from pathlib import Path

import pandas as pd

from muse import examples
from muse.wizard import add_agent, get_sectors

parent_path = Path(__file__).parent


def generate_model_1():
    """Generates the first model for tutorial 1.

    Adds a new agent to the model with a single objective.

    """
    model_name = "1-single-objective"
    model_path = parent_path / model_name
    if model_path.exists():
        shutil.rmtree(model_path)

    # Starting point: copy default model
    examples.copy_model(name="default", path=parent_path, overwrite=True)
    os.rename(parent_path / "model", model_path)

    # Copy agent A1 -> A2
    add_agent(
        model_path,
        agent_name="A2",
        copy_from="A1",
        agentshare_new="Agent2",
    )

    # Change objective for agent A2
    agents_file = model_path / "Agents.csv"
    df = pd.read_csv(agents_file)
    df.loc[df["name"] == "A2", "objective1"] = "fuel_consumption_cost"
    df.to_csv(agents_file, index=False)

    # Split population between the two agents
    agents_file = model_path / "Agents.csv"
    df = pd.read_csv(agents_file)
    df.loc[:, "quantity"] = 0.5
    df.to_csv(agents_file, index=False)

    # Split capacity equally between the two agents
    for sector in get_sectors(model_path):
        technodata_file = model_path / f"{sector}/Technodata.csv"
        df = pd.read_csv(technodata_file)
        df.loc[:, "Agent1"] = 0.5
        df.loc[:, "Agent2"] = 0.5
        df.to_csv(technodata_file, index=False)


def generate_model_2():
    """Generates the second model for tutorial 2.

    Adds a second objective for agent A2.

    """
    model_name = "2-multiple-objective"
    model_path = parent_path / model_name
    if model_path.exists():
        shutil.rmtree(model_path)

    # Starting point: copy model from previous tutorial
    shutil.copytree(parent_path / "1-single-objective", model_path)
    if (model_path / "Results").exists():
        shutil.rmtree(model_path / "Results")

    # Add second objective for agent A2
    agents_file = model_path / "Agents.csv"
    df = pd.read_csv(agents_file)
    df.loc[df["name"] == "A2", "objective2"] = "LCOE"
    df.loc[df["name"] == "A2", "decision_method"] = "weighted_sum"
    df.loc[df["name"] == "A2", ["obj_data1", "obj_data2"]] = 0.5
    df.loc[df["name"] == "A2", "obj_sort2"] = True
    df.to_csv(agents_file, index=False)


if __name__ == "__main__":
    generate_model_1()
    generate_model_2()



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any

from xarray import Dataset


class AbstractSector(ABC):
    """Abstract base class for sectors.

    Sectors are part of type hierarchy with :py:class:`AbstractSector` at the apex: all
    sectors should derive from :py:class:`AbstractSector` directly or indirectly.

    MUSE only requires two things of a sector. Sector should be instanstiable via a
    :py:meth:`~AbstractSector.factory` function. And they should be callable via
    :py:meth:`~AbstractSector.next`.

    :py:class:`AbstractSector` declares an interface with these two functions. Sectors
    which derive from it will be warned if either method is not implemented.
    """

    @classmethod
    @abstractmethod
    def factory(cls, name: str, settings: Any) -> AbstractSector:
        """Creates class from settings named-tuple."""
        pass

    @abstractmethod
    def next(self, mca_market: Dataset) -> Dataset:
        """Advance sector by one time period."""

    def __repr__(self):
        return f"<{self.name.title()} sector - object at {hex(id(self))}>"



import os
import shutil
from pathlib import Path

from muse import examples

parent_path = Path(__file__).parent


def generate_model_1() -> None:
    """Generates the first model for tutorial 6.

    Adds correlation demand for residential sector.

    """
    model_name = "1-correlation"
    model_path = parent_path / model_name
    if model_path.exists():
        shutil.rmtree(model_path)

    # Copy default_correlation model
    examples.copy_model(name="default_correlation", path=parent_path, overwrite=True)
    os.rename(parent_path / "model", model_path)


if __name__ == "__main__":
    generate_model_1()



import os
import shutil
from concurrent.futures import ProcessPoolExecutor
from pathlib import Path
from typing import Union

from muse.examples import AVAILABLE_EXAMPLES, model


def run_model(name: str) -> Union[str, None]:
    """Run a model with the specified settings file.

    Args:
        name: Name of the model to run

    Returns:
        Error message for failed runs, otherwise None.
    """
    try:
        # Prepare results folder
        folder_path = Path(name)
        if folder_path.exists():
            shutil.rmtree(folder_path)
        folder_path.mkdir()

        # Run model
        os.chdir(folder_path)
        m = model(name)
        m.run()
        return None
    except Exception as e:
        return f"Failed to run model {name}. Error: {e}"


if __name__ == "__main__":
    with ProcessPoolExecutor() as executor:
        futures = {executor.submit(run_model, model) for model in AVAILABLE_EXAMPLES}
    for future in futures:
        error_message = future.result()
        if error_message:
            print(error_message)



"""Functions to modify MUSE input files.

For example, adding new commodities or agents to a model.

These functions are designed specifically to work with models following the conventional
file structure and naming scheme of the example models provided with MUSE, and will
not necessarily work with models that deviate from this structure.
"""

from __future__ import annotations

from itertools import chain
from pathlib import Path
from typing import Callable

import pandas as pd
from tomlkit import dumps, parse


def modify_toml(path_to_toml: str | Path, function: Callable):
    """Apply the specified function to modify a toml file.

    Args:
        path_to_toml: Path to the toml file (string or Path object)
        function: Function to apply to the toml data. Must take a dictionary as a single
            input and modify it in place.

    """
    path_to_toml = Path(path_to_toml)
    data = parse(path_to_toml.read_text())
    function(data)
    path_to_toml.write_text(dumps(data))


def get_sectors(model_path: Path) -> list[str]:
    """Get a list of sector names for a model.

    Args:
        model_path: Path to the model folder.

    Returns:
        List of sector names
    """
    return [s.name for s in model_path.iterdir() if (s / "Technodata.csv").is_file()]


def add_new_commodity(
    model_path: Path, commodity_name: str, sector: str, copy_from: str
) -> None:
    """Add a new commodity to a sector by copying an existing one.

    Args:
        model_path: Path to the model folder.
        commodity_name: Name of the new commodity.
        sector: Sector to add the commodity to.
        copy_from: Name of the commodity to copy from.
    """
    # Add commodity to global commodities file
    global_commodities_file = model_path / "GlobalCommodities.csv"
    df = pd.read_csv(global_commodities_file)
    new_rows = df[df["commodity"] == copy_from].assign(commodity=commodity_name)
    df = pd.concat([df, new_rows])
    df.to_csv(global_commodities_file, index=False)


def add_new_process(
    model_path: Path, process_name: str, sector: str, copy_from: str
) -> None:
    """Add a new process to a sector by copying an existing one.

    Args:
        model_path: Path to the model folder.
        process_name: Name of the new process.
        sector: Sector to add the process to.
        copy_from: Name of the process to copy from.
    """
    files_to_update = (
        model_path / file
        for file in (
            f"{sector}/CommIn.csv",
            f"{sector}/CommOut.csv",
            f"{sector}/ExistingCapacity.csv",
            f"{sector}/Technodata.csv",
        )
    )

    for file in files_to_update:
        df = pd.read_csv(file)
        new_rows = df[df["technology"] == copy_from].copy()
        new_rows["technology"] = process_name
        df = pd.concat([df, new_rows])
        df.to_csv(file, index=False)


def add_technodata_for_new_year(
    model_path: Path, year: str, sector: str, copy_from: str
) -> None:
    """Add price data for a new year by copying from an existing year.

    Args:
        model_path: Path to the model folder.
        year: Year to add the price data to.
        sector: Sector to add the price data to.
        copy_from: Year to copy the price data from.
    """
    file_to_update = model_path / f"{sector}/Technodata.csv"

    df = pd.read_csv(file_to_update)
    df["index"] = df.index
    new_rows = df[df["year"] == copy_from].copy()
    new_rows["year"] = year
    df = pd.concat([df, new_rows], ignore_index=True)
    df.sort_values(by=["index", "year"], inplace=True)
    df.drop(columns=["index"], inplace=True)
    df.to_csv(file_to_update, index=False)


def add_agent(
    model_path: Path,
    agent_name: str,
    copy_from: str,
    agentshare_new: str | None = None,
    agentshare_retrofit: str | None = None,
) -> None:
    """Add a new agent to the model by copying an existing one.

    Args:
        model_path: Path to the model folder.
        agent_name: Name of the new agent.
        copy_from: Name of the agent to copy from.
        agentshare_new: Name of the 'new' agent share for new agent. If None, the new
            agent will not have a 'new' share.
        agentshare_retrofit: Name of the 'retrofit' agent share for new agent. If None,
            the new agent will not have a 'retrofit' share.
    """
    agents_file = model_path / "Agents.csv"
    agents_df = pd.read_csv(agents_file)

    # Create mapping between share names
    copy_to_shares = {"New": agentshare_new, "Retrofit": agentshare_retrofit}
    copy_from_shares = {}
    for share_type in ["New", "Retrofit"]:
        filtered_df = agents_df.loc[
            (agents_df["name"] == copy_from) & (agents_df["type"] == share_type),
            "agent_share",
        ]
        copy_from_shares[share_type] = (
            filtered_df.iat[0] if not filtered_df.empty else None
        )

    # Update agents file
    for share_type in ["New", "Retrofit"]:
        if copy_to_shares[share_type] and copy_from_shares[share_type]:
            rows = agents_df[
                (agents_df["name"] == copy_from)
                & (agents_df["agent_share"] == copy_from_shares[share_type])
            ].copy()
            rows["name"] = agent_name
            rows["agent_share"] = copy_to_shares[share_type]
            agents_df = pd.concat([agents_df, rows])
    agents_df.to_csv(agents_file, index=False)

    # Update technodata files for each sector
    for sector in get_sectors(model_path):
        technodata_file = model_path / f"{sector}/Technodata.csv"
        technodata_df = pd.read_csv(technodata_file)
        for share_type in ["New", "Retrofit"]:
            if (
                copy_to_shares[share_type]
                and copy_from_shares[share_type]
                and copy_from_shares[share_type] in technodata_df.columns
            ):
                technodata_df[copy_to_shares[share_type]] = technodata_df[
                    copy_from_shares[share_type]
                ]
        technodata_df.to_csv(technodata_file, index=False)


def add_region(model_path: Path, region_name: str, copy_from: str) -> None:
    """Add a new region to the MUSE model.

    Args:
        model_path: The path to the MUSE model directory.
        region_name: The name of the new region to be added.
        copy_from: The name of the region to copy data from.
    """
    # Append region to settings.toml
    settings_file = model_path / "settings.toml"
    modify_toml(settings_file, lambda x: x["regions"].append(region_name))

    # Modify csv files
    sector_files = (
        model_path / sector / file
        for sector in get_sectors(model_path)
        for file in (
            "Technodata.csv",
            "CommIn.csv",
            "CommOut.csv",
            "ExistingCapacity.csv",
        )
    )
    preset_files = model_path.glob("*preset*/*")
    global_files = (
        model_path / file
        for file in (
            "Agents.csv",
            "Projections.csv",
        )
    )
    for file_path in chain(sector_files, preset_files, global_files):
        df = pd.read_csv(file_path)
        new_rows = df[df["region"] == copy_from].copy()
        new_rows["region"] = region_name
        df = pd.concat([df, new_rows])
        df.to_csv(file_path, index=False)


def add_timeslice(model_path: Path, timeslice_name: str, copy_from: str) -> None:
    """Add a new timeslice to the model.

    Args:
        model_path: The path to the model directory.
        timeslice_name: The name of the new timeslice.
        copy_from: The name of the timeslice to copy from.
    """
    # Append timeslice to timeslices in settings.toml
    settings_file = model_path / "settings.toml"
    settings = parse(settings_file.read_text())
    timeslices = settings["timeslices"]["all-year"]["all-week"]
    copy_from_number = list(timeslices).index(copy_from) + 1
    timeslices[timeslice_name] = timeslices[copy_from]
    settings_file.write_text(dumps(settings))

    # Loop through all preset files
    preset_files = model_path.glob("*preset*/*")
    for file_path in preset_files:
        df = pd.read_csv(file_path)
        new_rows = df[df["timeslice"] == copy_from_number].copy()
        new_rows["timeslice"] = len(timeslices)
        df = pd.concat([df, new_rows])
        df = df.sort_values(by=["region", "timeslice"]).reset_index(drop=True)
        df.to_csv(file_path, index=False)
